"""Implement the :stac-ext:`View Geometry Extension <view>`."""

from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Generic, Literal, TypeVar, cast

import pystac
from pystac.extensions.base import (
    ExtensionManagementMixin,
    PropertiesExtension,
    SummariesExtension,
)
from pystac.extensions.hooks import ExtensionHooks
from pystac.summaries import RangeSummary

#: Generalized version of :class:`~pystac.Item`, :class:`~pystac.Asset`
#: or :class:`~pystac.ItemAssetDefinition`
T = TypeVar("T", pystac.Item, pystac.Asset, pystac.ItemAssetDefinition)

SCHEMA_URI: str = "https://stac-extensions.github.io/view/v1.0.0/schema.json"
PREFIX: str = "view:"

OFF_NADIR_PROP: str = PREFIX + "off_nadir"
INCIDENCE_ANGLE_PROP: str = PREFIX + "incidence_angle"
AZIMUTH_PROP: str = PREFIX + "azimuth"
SUN_AZIMUTH_PROP: str = PREFIX + "sun_azimuth"
SUN_ELEVATION_PROP: str = PREFIX + "sun_elevation"


class ViewExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """An abstract class that can be used to extend the properties of an
    :class:`~pystac.Item` with properties from the :stac-ext:`View Geometry
    Extension <view>`. This class is generic over the type of STAC Object to be
    extended (e.g. :class:`~pystac.Item`, :class:`~pystac.Asset`).

    To create a concrete instance of :class:`ViewExtension`, use the
    :meth:`ViewExtension.ext` method. For example:

    .. code-block:: python

       >>> item: pystac.Item = ...
       >>> view_ext = ViewExtension.ext(item)
    """

    name: Literal["view"] = "view"

    def apply(
        self,
        off_nadir: float | None = None,
        incidence_angle: float | None = None,
        azimuth: float | None = None,
        sun_azimuth: float | None = None,
        sun_elevation: float | None = None,
    ) -> None:
        """Applies View Geometry extension properties to the extended
        :class:`~pystac.Item`.

        Args:
            off_nadir : The angle from the sensor between nadir (straight down)
                and the scene center. Measured in degrees (0-90).
            incidence_angle : The incidence angle is the angle between the
                vertical (normal) to the intercepting surface and the line of sight
                back to the satellite at the scene center. Measured in degrees (0-90).
            azimuth : Viewing azimuth angle. The angle measured from the
                sub-satellite point (point on the ground below the platform) between
                the scene center and true north. Measured clockwise from north in
                degrees (0-360).
            sun_azimuth : Sun azimuth angle. From the scene center point on the
                ground, this is the angle between truth north and the sun. Measured
                clockwise in degrees (0-360).
            sun_elevation : Sun elevation angle. The angle from the tangent of
                the scene center point to the sun. Measured from the horizon in
                degrees (0-90).
        """
        self.off_nadir = off_nadir
        self.incidence_angle = incidence_angle
        self.azimuth = azimuth
        self.sun_azimuth = sun_azimuth
        self.sun_elevation = sun_elevation

    @property
    def off_nadir(self) -> float | None:
        """Get or sets the angle from the sensor between nadir (straight down)
        and the scene center. Measured in degrees (0-90).
        """
        return self._get_property(OFF_NADIR_PROP, float)

    @off_nadir.setter
    def off_nadir(self, v: float | None) -> None:
        self._set_property(OFF_NADIR_PROP, v)

    @property
    def incidence_angle(self) -> float | None:
        """Get or sets the incidence angle is the angle between the vertical (normal)
        to the intercepting surface and the line of sight back to the satellite at
        the scene center. Measured in degrees (0-90).
        """
        return self._get_property(INCIDENCE_ANGLE_PROP, float)

    @incidence_angle.setter
    def incidence_angle(self, v: float | None) -> None:
        self._set_property(INCIDENCE_ANGLE_PROP, v)

    @property
    def azimuth(self) -> float | None:
        """Get or sets the viewing azimuth angle.

        The angle measured from the sub-satellite
        point (point on the ground below the platform) between the scene center and true
        north. Measured clockwise from north in degrees (0-360).
        """
        return self._get_property(AZIMUTH_PROP, float)

    @azimuth.setter
    def azimuth(self, v: float | None) -> None:
        self._set_property(AZIMUTH_PROP, v)

    @property
    def sun_azimuth(self) -> float | None:
        """Get or sets the sun azimuth angle.

        From the scene center point on the ground, this
        is the angle between truth north and the sun. Measured clockwise in
        degrees (0-360).
        """
        return self._get_property(SUN_AZIMUTH_PROP, float)

    @sun_azimuth.setter
    def sun_azimuth(self, v: float | None) -> None:
        self._set_property(SUN_AZIMUTH_PROP, v)

    @property
    def sun_elevation(self) -> float | None:
        """Get or sets the sun elevation angle. The angle from the tangent of the scene
        center point to the sun. Measured from the horizon in degrees (0-90).
        """
        return self._get_property(SUN_ELEVATION_PROP, float)

    @sun_elevation.setter
    def sun_elevation(self, v: float | None) -> None:
        self._set_property(SUN_ELEVATION_PROP, v)

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> ViewExtension[T]:
        """Extends the given STAC Object with properties from the :stac-ext:`View
        Geometry Extension <scientific>`.

        This extension can be applied to instances of :class:`~pystac.Item` or
        :class:`~pystac.Asset`.

        Raises:

            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(ViewExtension[T], ItemViewExtension(obj))
        elif isinstance(obj, pystac.Asset):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(ViewExtension[T], AssetViewExtension(obj))
        elif isinstance(obj, pystac.ItemAssetDefinition):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(ViewExtension[T], ItemAssetsViewExtension(obj))
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))

    @classmethod
    def summaries(
        cls, obj: pystac.Collection, add_if_missing: bool = False
    ) -> SummariesViewExtension:
        """Returns the extended summaries object for the given collection."""
        cls.ensure_has_extension(obj, add_if_missing)
        return SummariesViewExtension(obj)


class ItemViewExtension(ViewExtension[pystac.Item]):
    """A concrete implementation of :class:`ViewExtension` on an :class:`~pystac.Item`
    that extends the properties of the Item to include properties defined in the
    :stac-ext:`View Geometry Extension <view>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`ViewExtension.ext` on an :class:`~pystac.Item` to extend it.
    """

    item: pystac.Item
    """The :class:`~pystac.Item` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Item` properties, including extension properties."""

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    def __repr__(self) -> str:
        return f"<ItemViewExtension Item id={self.item.id}>"


class AssetViewExtension(ViewExtension[pystac.Asset]):
    """A concrete implementation of :class:`ViewExtension` on an :class:`~pystac.Asset`
    that extends the Asset fields to include properties defined in the
    :stac-ext:`View Geometry Extension <view>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`ViewExtension.ext` on an :class:`~pystac.Asset` to extend it.
    """

    asset_href: str
    """The ``href`` value of the :class:`~pystac.Asset` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Asset` fields, including extension properties."""

    additional_read_properties: Iterable[dict[str, Any]] | None = None
    """If present, this will be a list containing 1 dictionary representing the
    properties of the owning :class:`~pystac.Item`."""

    def __init__(self, asset: pystac.Asset):
        self.asset_href = asset.href
        self.properties = asset.extra_fields
        if asset.owner and isinstance(asset.owner, pystac.Item):
            self.additional_read_properties = [asset.owner.properties]

    def __repr__(self) -> str:
        return f"<AssetViewExtension Asset href={self.asset_href}>"


class ItemAssetsViewExtension(ViewExtension[pystac.ItemAssetDefinition]):
    properties: dict[str, Any]
    asset_defn: pystac.ItemAssetDefinition

    def __init__(self, item_asset: pystac.ItemAssetDefinition):
        self.asset_defn = item_asset
        self.properties = item_asset.properties


class SummariesViewExtension(SummariesExtension):
    """A concrete implementation of :class:`~pystac.extensions.base.SummariesExtension`
    that extends the ``summaries`` field of a :class:`~pystac.Collection` to include
    properties defined in the :stac-ext:`View Object Extension <view>`.
    """

    @property
    def off_nadir(self) -> RangeSummary[float] | None:
        """Get or sets the summary of :attr:`ViewExtension.off_nadir` values for
        this Collection.
        """
        return self.summaries.get_range(OFF_NADIR_PROP)

    @off_nadir.setter
    def off_nadir(self, v: RangeSummary[float] | None) -> None:
        self._set_summary(OFF_NADIR_PROP, v)

    @property
    def incidence_angle(self) -> RangeSummary[float] | None:
        """Get or sets the summary of :attr:`ViewExtension.incidence_angle` values
        for this Collection.
        """
        return self.summaries.get_range(INCIDENCE_ANGLE_PROP)

    @incidence_angle.setter
    def incidence_angle(self, v: RangeSummary[float] | None) -> None:
        self._set_summary(INCIDENCE_ANGLE_PROP, v)

    @property
    def azimuth(self) -> RangeSummary[float] | None:
        """Get or sets the summary of :attr:`ViewExtension.azimuth` values
        for this Collection.
        """
        return self.summaries.get_range(AZIMUTH_PROP)

    @azimuth.setter
    def azimuth(self, v: RangeSummary[float] | None) -> None:
        self._set_summary(AZIMUTH_PROP, v)

    @property
    def sun_azimuth(self) -> RangeSummary[float] | None:
        """Get or sets the summary of :attr:`ViewExtension.sun_azimuth` values
        for this Collection.
        """
        return self.summaries.get_range(SUN_AZIMUTH_PROP)

    @sun_azimuth.setter
    def sun_azimuth(self, v: RangeSummary[float] | None) -> None:
        self._set_summary(SUN_AZIMUTH_PROP, v)

    @property
    def sun_elevation(self) -> RangeSummary[float] | None:
        """Get or sets the summary of :attr:`ViewExtension.sun_elevation` values
        for this Collection.
        """
        return self.summaries.get_range(SUN_ELEVATION_PROP)

    @sun_elevation.setter
    def sun_elevation(self, v: RangeSummary[float] | None) -> None:
        self._set_summary(SUN_ELEVATION_PROP, v)


class ViewExtensionHooks(ExtensionHooks):
    schema_uri = SCHEMA_URI
    prev_extension_ids = {"view"}
    stac_object_types = {pystac.STACObjectType.ITEM}


VIEW_EXTENSION_HOOKS: ExtensionHooks = ViewExtensionHooks()



import json

import pytest

import pystac
from pystac import ExtensionTypeError, Item
from pystac.extensions.raster import (
    DataType,
    Histogram,
    NoDataStrings,
    RasterBand,
    RasterExtension,
    Sampling,
    Statistics,
)
from pystac.utils import get_opt
from tests.conftest import get_data_file
from tests.utils import TestCases, assert_to_from_dict

PLANET_EXAMPLE_URI = get_data_file("raster/raster-planet-example.json")


@pytest.fixture
def ext_item() -> pystac.Item:
    return pystac.Item.from_file(PLANET_EXAMPLE_URI)


SENTINEL2_EXAMPLE_URI = TestCases.get_path(
    "data-files/raster/raster-sentinel2-example.json"
)


LANDSAT_COLLECTION_EXAMPLE_URI = TestCases.get_path(
    "data-files/raster/landsat-collection-example.json"
)


def test_to_from_dict() -> None:
    with open(PLANET_EXAMPLE_URI) as f:
        item_dict = json.load(f)
    assert_to_from_dict(Item, item_dict)


@pytest.mark.vcr()
def test_validate_raster(ext_item: pystac.Item) -> None:
    item2 = pystac.Item.from_file(SENTINEL2_EXAMPLE_URI)
    ext_item.validate()
    item2.validate()


@pytest.mark.vcr()
def test_asset_bands(ext_item: pystac.Item) -> None:
    item2 = pystac.Item.from_file(SENTINEL2_EXAMPLE_URI)

    # Get
    data_asset = ext_item.assets["data"]
    asset_bands = RasterExtension.ext(data_asset).bands
    assert asset_bands is not None
    assert len(asset_bands) == 4
    assert asset_bands[0].nodata == 0
    assert asset_bands[0].sampling == Sampling.AREA
    assert asset_bands[0].unit == "W⋅sr−1⋅m−2⋅nm−1"
    assert asset_bands[0].data_type == DataType.UINT16
    assert asset_bands[0].scale == 0.01
    assert asset_bands[0].offset == 0
    assert asset_bands[0].spatial_resolution == 3

    band0_stats = asset_bands[0].statistics
    assert band0_stats is not None
    assert band0_stats.minimum == 1962
    assert band0_stats.maximum == 32925
    assert band0_stats.mean == 8498.9400644319
    assert band0_stats.stddev == 5056.1292002722
    assert band0_stats.valid_percent == 61.09

    band0_hist = asset_bands[0].histogram
    assert band0_hist is not None
    assert band0_hist.count == 256
    assert band0_hist.min == 1901.288235294118
    assert band0_hist.max == 32985.71176470588
    assert len(band0_hist.buckets) == band0_hist.count

    index_asset = ext_item.assets["metadata"]
    asset_bands = RasterExtension.ext(index_asset).bands
    assert None is asset_bands

    b09_asset = item2.assets["B09"]
    b09_bands = RasterExtension.ext(b09_asset).bands
    assert b09_bands is not None
    assert b09_bands[0].nodata == "nan"

    # Set
    b2_asset = item2.assets["B02"]
    assert (
        get_opt(get_opt(RasterExtension.ext(b2_asset).bands)[0].statistics).maximum
        == 19264
    )
    b1_asset = item2.assets["B01"]
    RasterExtension.ext(b2_asset).bands = RasterExtension.ext(b1_asset).bands

    new_b2_asset_bands = RasterExtension.ext(item2.assets["B02"]).bands

    assert get_opt(get_opt(new_b2_asset_bands)[0].statistics).maximum == 20567
    new_b2_asset_band0 = get_opt(new_b2_asset_bands)[0]
    new_b2_asset_band0.nodata = NoDataStrings.INF

    item2.validate()

    # Check adding a new asset
    new_stats = [
        Statistics.create(
            minimum=0, maximum=10000, mean=5000, stddev=10, valid_percent=88
        ),
        Statistics.create(minimum=-1, maximum=1, mean=0, stddev=1, valid_percent=100),
        Statistics.create(
            minimum=1, maximum=255, mean=200, stddev=3, valid_percent=100
        ),
    ]
    # new_histograms = []
    with open(TestCases.get_path("data-files/raster/gdalinfo.json")) as gdaljson_file:
        gdaljson_data = json.load(gdaljson_file)
        new_histograms = list(
            map(
                lambda band: Histogram.from_dict(band["histogram"]),
                gdaljson_data["bands"],
            )
        )
    new_bands = [
        RasterBand.create(
            nodata=1,
            unit="test1",
            statistics=new_stats[0],
            histogram=new_histograms[0],
        ),
        RasterBand.create(
            nodata=2,
            unit="test2",
            statistics=new_stats[1],
            histogram=new_histograms[1],
        ),
        RasterBand.create(
            nodata=NoDataStrings.NINF,
            unit="test3",
            statistics=new_stats[2],
            histogram=new_histograms[2],
        ),
    ]
    asset = pystac.Asset(href="some/path.tif", media_type=pystac.MediaType.GEOTIFF)
    RasterExtension.ext(asset).bands = new_bands
    ext_item.add_asset("test", asset)

    assert len(ext_item.assets["test"].extra_fields["raster:bands"]) == 3
    assert (
        ext_item.assets["test"].extra_fields["raster:bands"][1]["statistics"]["minimum"]
        == -1
    )
    assert (
        ext_item.assets["test"].extra_fields["raster:bands"][1]["histogram"]["min"]
        == 3848.354901960784
    )
    assert ext_item.assets["test"].extra_fields["raster:bands"][2]["nodata"] == "-inf"

    for s in new_stats:
        s.minimum = None
        s.maximum = None
        s.mean = None
        s.stddev = None
        s.valid_percent = None
        assert len(s.properties) == 0

    for b in new_bands:
        b.bits_per_sample = None
        b.data_type = None
        b.histogram = None
        b.nodata = None
        b.sampling = None
        b.scale = None
        b.spatial_resolution = None
        b.statistics = None
        b.unit = None
        b.offset = None
        assert len(b.properties) == 0

    new_stats[2].apply(minimum=0, maximum=10000, mean=5000, stddev=10, valid_percent=88)
    new_stats[1].apply(minimum=-1, maximum=1, mean=0, stddev=1, valid_percent=100)
    new_stats[0].apply(minimum=1, maximum=255, mean=200, stddev=3, valid_percent=100)
    new_bands[2].apply(
        nodata=1,
        unit="test1",
        statistics=new_stats[2],
        histogram=new_histograms[0],
    )
    new_bands[1].apply(
        nodata=2,
        unit="test2",
        statistics=new_stats[1],
        histogram=new_histograms[1],
    )
    new_bands[0].apply(
        nodata=NoDataStrings.NAN,
        unit="test3",
        statistics=new_stats[0],
        histogram=new_histograms[2],
    )
    RasterExtension.ext(ext_item.assets["test"]).apply(new_bands)
    assert (
        ext_item.assets["test"].extra_fields["raster:bands"][0]["statistics"]["minimum"]
        == 1
    )
    assert ext_item.assets["test"].extra_fields["raster:bands"][0]["nodata"] == "nan"


def test_extension_not_implemented(ext_item: pystac.Item) -> None:
    # Should raise exception if Item does not include extension URI
    ext_item.stac_extensions.remove(RasterExtension.get_schema_uri())

    # Should raise exception if owning Item does not include extension URI
    asset = ext_item.assets["data"]

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = RasterExtension.ext(asset)

    # Should succeed if Asset has no owner
    ownerless_asset = pystac.Asset.from_dict(asset.to_dict())
    _ = RasterExtension.ext(ownerless_asset)


def test_ext_add_to(ext_item: pystac.Item) -> None:
    ext_item.stac_extensions.remove(RasterExtension.get_schema_uri())
    asset = ext_item.assets["data"]

    _ = RasterExtension.ext(asset, add_if_missing=True)

    assert RasterExtension.get_schema_uri() in ext_item.stac_extensions


def test_should_raise_exception_when_passing_invalid_extension_object() -> None:
    with pytest.raises(
        ExtensionTypeError, match=r"^RasterExtension does not apply to type 'object'$"
    ):
        # calling it wrong on purpose so --------v
        RasterExtension.ext(object())  # type: ignore


def test_summaries_adds_uri() -> None:
    col = pystac.Collection.from_file(LANDSAT_COLLECTION_EXAMPLE_URI)
    col.stac_extensions = []
    with pytest.raises(
        pystac.ExtensionNotImplemented,
        match="Extension 'raster' is not implemented",
    ):
        RasterExtension.summaries(col, add_if_missing=False)

    RasterExtension.summaries(col, True)

    assert RasterExtension.get_schema_uri() in col.stac_extensions

    RasterExtension.remove_from(col)
    assert RasterExtension.get_schema_uri() not in col.stac_extensions


def test_collection_item_asset() -> None:
    coll = pystac.Collection.from_file(LANDSAT_COLLECTION_EXAMPLE_URI)

    assert coll.item_assets

    qa = coll.item_assets["qa"]
    ang = coll.item_assets["ang"]

    assert qa.ext.raster.bands is not None
    assert ang.ext.raster.bands is None


def test_older_extension_version(ext_item: pystac.Item) -> None:
    old = "https://stac-extensions.github.io/raster/v1.0.0/schema.json"
    new = "https://stac-extensions.github.io/raster/v1.1.0/schema.json"

    stac_extensions = set(ext_item.stac_extensions)
    stac_extensions.remove(new)
    stac_extensions.add(old)
    item_as_dict = ext_item.to_dict(include_self_link=False, transform_hrefs=False)
    item_as_dict["stac_extensions"] = list(stac_extensions)
    item = pystac.Item.from_dict(item_as_dict, migrate=False)
    assert RasterExtension.has_extension(item)
    assert old in item.stac_extensions

    migrated_item = pystac.Item.from_dict(item_as_dict, migrate=True)
    assert RasterExtension.has_extension(migrated_item)
    assert new in migrated_item.stac_extensions



import posixpath
from collections.abc import Callable
from datetime import datetime, timedelta

import pytest

import pystac
from pystac.collection import Collection
from pystac.layout import (
    APILayoutStrategy,
    AsIsLayoutStrategy,
    BestPracticesLayoutStrategy,
    CustomLayoutStrategy,
    LayoutTemplate,
    TemplateLayoutStrategy,
)
from tests.utils import (
    ARBITRARY_BBOX,
    ARBITRARY_GEOM,
    TestCases,
    path_includes_drive_letter,
)


class TestLayoutTemplate:
    def test_templates_item_datetime(self) -> None:
        year = 2020
        month = 11
        day = 3
        date = "2020-11-03"
        dt = datetime(year, month, day, 18, 30)

        template = LayoutTemplate("${year}/${month}/${day}/${date}/item.json")

        item = pystac.Item(
            "test",
            geometry=ARBITRARY_GEOM,
            bbox=ARBITRARY_BBOX,
            datetime=dt,
            properties={},
        )

        parts = template.get_template_values(item)

        assert set(parts) == {"year", "month", "day", "date"}

        assert parts["year"] == year
        assert parts["month"] == month
        assert parts["day"] == day
        assert parts["date"] == date

        path = template.substitute(item)
        assert path == "2020/11/3/2020-11-03/item.json"

    def test_templates_item_start_datetime(self) -> None:
        year = 2020
        month = 11
        day = 3
        date = "2020-11-03"
        dt = datetime(year, month, day, 18, 30)

        template = LayoutTemplate("${year}/${month}/${day}/${date}/item.json")

        item = pystac.Item(
            "test",
            geometry=ARBITRARY_GEOM,
            bbox=ARBITRARY_BBOX,
            datetime=None,
            properties={
                "start_datetime": dt.isoformat(),
                "end_datetime": (dt + timedelta(days=1)).isoformat(),
            },
        )

        parts = template.get_template_values(item)

        assert set(parts) == {"year", "month", "day", "date"}

        assert parts["year"] == year
        assert parts["month"] == month
        assert parts["day"] == day
        assert parts["date"] == date

        path = template.substitute(item)
        assert path == "2020/11/3/2020-11-03/item.json"

    def test_templates_item_collection(self) -> None:
        template = LayoutTemplate("${collection}/item.json")

        collection = TestCases.case_4().get_child("acc")
        assert collection is not None
        item = next(collection.get_items())
        assert item.collection_id is not None

        parts = template.get_template_values(item)
        assert len(parts) == 1
        assert "collection" in parts
        assert parts["collection"] == item.collection_id

        path = template.substitute(item)
        assert path == f"{item.collection_id}/item.json"

    def test_throws_for_no_collection(self) -> None:
        template = LayoutTemplate("${collection}/item.json")

        collection = TestCases.case_4().get_child("acc")
        assert collection is not None
        item = next(collection.get_items())
        item.set_collection(None)
        assert item.collection_id is None

        with pytest.raises(pystac.TemplateError):
            template.get_template_values(item)

    def test_nested_properties(self) -> None:
        dt = datetime(2020, 11, 3, 18, 30)

        template = LayoutTemplate("${test.prop}/${ext:extra.test.prop}/item.json")

        item = pystac.Item(
            "test",
            geometry=ARBITRARY_GEOM,
            bbox=ARBITRARY_BBOX,
            datetime=dt,
            properties={"test": {"prop": 4326}},
            extra_fields={"ext:extra": {"test": {"prop": 3857}}},
        )

        parts = template.get_template_values(item)

        assert set(parts) == {"test.prop", "ext:extra.test.prop"}

        assert parts["test.prop"] == 4326
        assert parts["ext:extra.test.prop"] == 3857

        path = template.substitute(item)

        assert path == "4326/3857/item.json"

    def test_substitute_with_colon_properties(self) -> None:
        dt = datetime(2020, 11, 3, 18, 30)

        template = LayoutTemplate("${ext:prop}/item.json")

        item = pystac.Item(
            "test",
            geometry=ARBITRARY_GEOM,
            bbox=ARBITRARY_BBOX,
            datetime=dt,
            properties={"ext:prop": 1},
        )

        path = template.substitute(item)

        assert path == "1/item.json"

    def test_defaults(self) -> None:
        template = LayoutTemplate(
            "${doesnotexist}/collection.json", defaults={"doesnotexist": "yes"}
        )

        catalog = TestCases.case_4().get_child("acc")
        assert catalog is not None
        catalog.extra_fields = {"one": "two"}
        path = template.substitute(catalog)

        assert path == "yes/collection.json"

    def test_docstring_examples(self) -> None:
        item = pystac.Item.from_file(
            TestCases.get_path(
                "data-files/examples/1.0.0-beta.2/item-spec/"
                "examples/landsat8-sample.json"
            )
        )
        item.common_metadata.license = "CC-BY-3.0"
        # Uses the year, month and day of the item
        template1 = LayoutTemplate("${year}/${month}/${day}")
        path1 = template1.substitute(item)
        assert path1 == "2014/6/2"

        # Uses a custom extension properties found on in the item properties.
        template2 = LayoutTemplate("${landsat:path}/${landsat:row}")
        path2 = template2.substitute(item)
        assert path2 == "107/18"

        # Uses the collection ID and a common metadata property for an item.
        template3 = LayoutTemplate("${collection}/${common_metadata.license}")
        path3 = template3.substitute(item)
        assert path3 == "landsat-8-l1/CC-BY-3.0"


class TestCustomLayoutStrategy:
    def get_custom_catalog_func(self) -> Callable[[pystac.Catalog, str, bool], str]:
        def fn(cat: pystac.Catalog, parent_dir: str, is_root: bool) -> str:
            return posixpath.join(parent_dir, f"cat/{is_root}/{cat.id}.json")

        return fn

    def get_custom_collection_func(
        self,
    ) -> Callable[[pystac.Collection, str, bool], str]:
        def fn(col: pystac.Collection, parent_dir: str, is_root: bool) -> str:
            return posixpath.join(parent_dir, f"col/{is_root}/{col.id}.json")

        return fn

    def get_custom_item_func(self) -> Callable[[pystac.Item, str], str]:
        def fn(item: pystac.Item, parent_dir: str) -> str:
            return posixpath.join(parent_dir, f"item/{item.id}.json")

        return fn

    def test_produces_layout_for_catalog(self) -> None:
        strategy = CustomLayoutStrategy(catalog_func=self.get_custom_catalog_func())
        cat = pystac.Catalog(id="test", description="test desc")
        href = strategy.get_href(cat, parent_dir="http://example.com", is_root=True)
        assert href == "http://example.com/cat/True/test.json"

    def test_produces_fallback_layout_for_catalog(self) -> None:
        fallback = BestPracticesLayoutStrategy()
        strategy = CustomLayoutStrategy(
            collection_func=self.get_custom_collection_func(),
            item_func=self.get_custom_item_func(),
            fallback_strategy=fallback,
        )
        cat = pystac.Catalog(id="test", description="test desc")
        href = strategy.get_href(cat, parent_dir="http://example.com")
        expected = fallback.get_href(cat, parent_dir="http://example.com")
        assert href == expected

    def test_produces_layout_for_collection(self) -> None:
        strategy = CustomLayoutStrategy(
            collection_func=self.get_custom_collection_func()
        )
        collection = TestCases.case_8()
        href = strategy.get_href(collection, parent_dir="http://example.com")
        assert href == f"http://example.com/col/False/{collection.id}.json"

    def test_produces_fallback_layout_for_collection(self) -> None:
        fallback = BestPracticesLayoutStrategy()
        strategy = CustomLayoutStrategy(
            catalog_func=self.get_custom_catalog_func(),
            item_func=self.get_custom_item_func(),
            fallback_strategy=fallback,
        )
        collection = TestCases.case_8()
        href = strategy.get_href(collection, parent_dir="http://example.com")
        expected = fallback.get_href(collection, parent_dir="http://example.com")
        assert href == expected

    def test_produces_layout_for_item(self) -> None:
        strategy = CustomLayoutStrategy(item_func=self.get_custom_item_func())
        collection = TestCases.case_8()
        item = next(collection.get_items(recursive=True))
        href = strategy.get_href(item, parent_dir="http://example.com")
        assert href == f"http://example.com/item/{item.id}.json"

    def test_produces_fallback_layout_for_item(self) -> None:
        fallback = BestPracticesLayoutStrategy()
        strategy = CustomLayoutStrategy(
            catalog_func=self.get_custom_catalog_func(),
            collection_func=self.get_custom_collection_func(),
            fallback_strategy=fallback,
        )
        collection = TestCases.case_8()
        item = next(collection.get_items(recursive=True))
        href = strategy.get_href(item, parent_dir="http://example.com")
        expected = fallback.get_href(item, parent_dir="http://example.com")
        assert href == expected


class TestTemplateLayoutStrategy:
    TEST_CATALOG_TEMPLATE = "cat/${id}/${description}"
    TEST_COLLECTION_TEMPLATE = "col/${id}/${license}"
    TEST_ITEM_TEMPLATE = "item/${collection}/${id}.json"

    def _get_collection(self) -> Collection:
        result = TestCases.case_4().get_child("acc")
        assert isinstance(result, Collection)
        return result

    def test_produces_layout_for_catalog(self) -> None:
        strategy = TemplateLayoutStrategy(catalog_template=self.TEST_CATALOG_TEMPLATE)
        cat = pystac.Catalog(id="test", description="test-desc")
        href = strategy.get_href(cat, parent_dir="http://example.com")
        assert href == "http://example.com/cat/test/test-desc/catalog.json"

    def test_produces_layout_for_catalog_with_filename(self) -> None:
        template = "cat/${id}/${description}/${id}.json"
        strategy = TemplateLayoutStrategy(catalog_template=template)
        cat = pystac.Catalog(id="test", description="test-desc")
        href = strategy.get_href(cat, parent_dir="http://example.com")
        assert href == "http://example.com/cat/test/test-desc/test.json"

    def test_produces_fallback_layout_for_catalog(self) -> None:
        fallback = BestPracticesLayoutStrategy()
        strategy = TemplateLayoutStrategy(
            collection_template=self.TEST_COLLECTION_TEMPLATE,
            item_template=self.TEST_ITEM_TEMPLATE,
            fallback_strategy=fallback,
        )
        cat = pystac.Catalog(id="test", description="test desc")
        href = strategy.get_href(cat, parent_dir="http://example.com")
        expected = fallback.get_href(cat, parent_dir="http://example.com")
        assert href == expected

    def test_produces_layout_for_collection(self) -> None:
        strategy = TemplateLayoutStrategy(
            collection_template=self.TEST_COLLECTION_TEMPLATE
        )
        collection = self._get_collection()
        href = strategy.get_href(collection, parent_dir="http://example.com")
        assert href == "http://example.com/col/{}/{}/collection.json".format(
            collection.id, collection.license
        )

    def test_produces_layout_for_collection_with_filename(self) -> None:
        template = "col/${id}/${license}/col.json"
        strategy = TemplateLayoutStrategy(collection_template=template)
        collection = self._get_collection()
        href = strategy.get_href(collection, parent_dir="http://example.com")
        assert href == "http://example.com/col/{}/{}/col.json".format(
            collection.id, collection.license
        )

    def test_produces_fallback_layout_for_collection(self) -> None:
        fallback = BestPracticesLayoutStrategy()
        strategy = TemplateLayoutStrategy(
            catalog_template=self.TEST_CATALOG_TEMPLATE,
            item_template=self.TEST_ITEM_TEMPLATE,
            fallback_strategy=fallback,
        )
        collection = self._get_collection()
        href = strategy.get_href(collection, parent_dir="http://example.com")
        expected = fallback.get_href(collection, parent_dir="http://example.com")
        assert href == expected

    def test_produces_layout_for_item(self) -> None:
        strategy = TemplateLayoutStrategy(item_template=self.TEST_ITEM_TEMPLATE)
        collection = self._get_collection()
        item = next(collection.get_items())
        href = strategy.get_href(item, parent_dir="http://example.com")
        assert href == f"http://example.com/item/{item.collection_id}/{item.id}.json"

    def test_produces_layout_for_item_without_filename(self) -> None:
        template = "item/${collection}"
        strategy = TemplateLayoutStrategy(item_template=template)
        collection = self._get_collection()
        item = next(collection.get_items())
        href = strategy.get_href(item, parent_dir="http://example.com")
        assert href == f"http://example.com/item/{item.collection_id}/{item.id}.json"

    def test_produces_fallback_layout_for_item(self) -> None:
        fallback = BestPracticesLayoutStrategy()
        strategy = TemplateLayoutStrategy(
            catalog_template=self.TEST_CATALOG_TEMPLATE,
            collection_template=self.TEST_COLLECTION_TEMPLATE,
            fallback_strategy=fallback,
        )
        collection = self._get_collection()
        item = next(collection.get_items())
        href = strategy.get_href(item, parent_dir="http://example.com")
        expected = fallback.get_href(item, parent_dir="http://example.com")
        assert href == expected


class TestBestPracticesLayoutStrategy:
    def test_produces_layout_for_root_catalog(self) -> None:
        strategy = BestPracticesLayoutStrategy()
        cat = pystac.Catalog(id="test", description="test desc")
        href = strategy.get_href(cat, parent_dir="http://example.com", is_root=True)
        assert href == "http://example.com/catalog.json"

    def test_produces_layout_for_child_catalog(self) -> None:
        strategy = BestPracticesLayoutStrategy()
        cat = pystac.Catalog(id="test", description="test desc")
        href = strategy.get_href(cat, parent_dir="http://example.com")
        assert href == "http://example.com/test/catalog.json"

    def test_produces_layout_for_root_collection(self) -> None:
        strategy = BestPracticesLayoutStrategy()
        collection = TestCases.case_8()
        href = strategy.get_href(
            collection, parent_dir="http://example.com", is_root=True
        )
        assert href == "http://example.com/collection.json"

    def test_produces_layout_for_child_collection(self) -> None:
        strategy = BestPracticesLayoutStrategy()
        collection = TestCases.case_8()
        href = strategy.get_href(collection, parent_dir="http://example.com")
        assert href == f"http://example.com/{collection.id}/collection.json"

    def test_produces_layout_for_item(self) -> None:
        strategy = BestPracticesLayoutStrategy()
        collection = TestCases.case_8()
        item = next(collection.get_items(recursive=True))
        href = strategy.get_href(item, parent_dir="http://example.com")
        expected = f"http://example.com/{item.id}/{item.id}.json"
        assert href == expected


class TestAsIsLayoutStrategy:
    def test_catalog(self) -> None:
        strategy = AsIsLayoutStrategy()
        expected_local_href = (
            "/an/href" if not path_includes_drive_letter() else "D:/an/href"
        )
        cat = pystac.Catalog(id="test", description="test desc")
        with pytest.raises(ValueError):
            strategy.get_href(cat, parent_dir="http://example.com", is_root=True)
        cat.set_self_href("/an/href")
        href = strategy.get_href(cat, parent_dir="https://example.com", is_root=True)
        assert href == expected_local_href

    def test_collection(self) -> None:
        strategy = AsIsLayoutStrategy()
        expected_local_href = (
            "/an/href" if not path_includes_drive_letter() else "D:/an/href"
        )
        collection = TestCases.case_8()
        collection.set_self_href(None)
        with pytest.raises(ValueError):
            strategy.get_href(collection, parent_dir="http://example.com", is_root=True)
        collection.set_self_href("/an/href")
        href = strategy.get_href(
            collection, parent_dir="https://example.com", is_root=True
        )
        assert href == expected_local_href

    def test_item(self) -> None:
        strategy = AsIsLayoutStrategy()
        expected_local_href = (
            "/an/href" if not path_includes_drive_letter() else "D:/an/href"
        )
        collection = TestCases.case_8()
        item = next(collection.get_items(recursive=True))
        item.set_self_href(None)
        with pytest.raises(ValueError):
            strategy.get_href(item, parent_dir="http://example.com")
        item.set_self_href("/an/href")
        href = strategy.get_href(item, parent_dir="http://example.com")
        assert href == expected_local_href


class TestAPILayoutStrategy:
    def test_produces_layout_for_root_catalog(self) -> None:
        strategy = APILayoutStrategy()
        cat = pystac.Catalog(id="test", description="test desc")
        href = strategy.get_href(cat, parent_dir="http://example.com", is_root=True)
        assert href == "http://example.com"

    def test_produces_layout_for_root_catalog_str(self) -> None:
        strategy = APILayoutStrategy()
        cat = pystac.Catalog(id="test", description="test desc")
        href = strategy.get_catalog_href(
            cat.id, parent_dir="http://example.com", is_root=True
        )
        assert href == "http://example.com"

    def test_produces_layout_for_child_catalog(self) -> None:
        strategy = APILayoutStrategy()
        cat = pystac.Catalog(id="test", description="test desc")
        href = strategy.get_href(cat, parent_dir="http://example.com")
        assert href == "http://example.com/test"

    def test_produces_layout_for_child_catalog_str(self) -> None:
        strategy = APILayoutStrategy()
        cat = pystac.Catalog(id="test", description="test desc")
        href = strategy.get_catalog_href(
            cat.id, parent_dir="http://example.com", is_root=False
        )
        assert href == "http://example.com/test"

    def test_cannot_produce_layout_for_root_collection(self) -> None:
        strategy = APILayoutStrategy()
        collection = TestCases.case_8()
        with pytest.raises(ValueError):
            strategy.get_href(collection, parent_dir="http://example.com", is_root=True)

    def test_produces_layout_for_child_collection(self) -> None:
        strategy = APILayoutStrategy()
        collection = TestCases.case_8()
        href = strategy.get_href(collection, parent_dir="http://example.com")
        assert href == f"http://example.com/collections/{collection.id}"

    def test_produces_layout_for_child_collection_str(self) -> None:
        strategy = APILayoutStrategy()
        collection = TestCases.case_8()
        href = strategy.get_collection_href(
            collection.id, parent_dir="http://example.com", is_root=False
        )
        assert href == f"http://example.com/collections/{collection.id}"

    def test_produces_layout_for_item(self) -> None:
        strategy = APILayoutStrategy()
        collection = TestCases.case_8()
        col_href = strategy.get_href(collection, parent_dir="http://example.com")
        item = next(collection.get_items(recursive=True))
        href = strategy.get_href(item, parent_dir=col_href)
        expected = f"http://example.com/collections/{collection.id}/items/{item.id}"
        assert href == expected

    def test_produces_layout_for_item_str(self) -> None:
        strategy = APILayoutStrategy()
        collection = TestCases.case_8()
        col_href = strategy.get_href(collection, parent_dir="http://example.com")
        item = next(collection.get_items(recursive=True))
        href = strategy.get_item_href(item.id, parent_dir=col_href)
        expected = f"http://example.com/collections/{collection.id}/items/{item.id}"
        assert href == expected

    def test_produces_normalized_layout(self) -> None:
        strategy = APILayoutStrategy()
        cat = pystac.Catalog(id="test_catalog", description="Test Catalog")
        col = pystac.Collection(
            id="test_collection",
            description="Test Collection",
            extent=pystac.Extent(
                spatial=pystac.SpatialExtent([[-180.0, -90.0, 180.0, 90.0]]),
                temporal=pystac.TemporalExtent(
                    [[datetime(2023, 1, 1), datetime(2023, 12, 31)]]
                ),
            ),
        )
        item = pystac.Item(
            id="test_item",
            geometry={
                "type": "Polygon",
                "coordinates": [
                    [
                        [180.0, -90.0],
                        [180.0, 90.0],
                        [-180.0, 90.0],
                        [-180.0, -90.0],
                        [180.0, -90.0],
                    ]
                ],
            },
            bbox=[-180, -90, 180, 90],
            datetime=datetime(2023, 1, 1),
            properties={},
            assets={
                "data": pystac.Asset(
                    href="http://example.com/assets/data.tif",
                    roles=["data"],
                    title="DATA",
                )
            },
        )
        cat.add_child(col)
        col.add_item(item)
        cat.normalize_hrefs("http://example.com", strategy=strategy)

        assert cat.self_href == "http://example.com"
        assert col.self_href == "http://example.com/collections/test_collection"
        assert (
            item.self_href
            == "http://example.com/collections/test_collection/items/test_item"
        )

    def test_produces_layout_for_search(self) -> None:
        strategy = APILayoutStrategy()
        href = strategy.get_search_href(parent_dir="http://example.com")
        assert href == "http://example.com/search"

    def test_produces_layout_for_conformance(self) -> None:
        strategy = APILayoutStrategy()
        href = strategy.get_conformance_href(parent_dir="http://example.com")
        assert href == "http://example.com/conformance"

    def test_produces_layout_for_service_description(self) -> None:
        strategy = APILayoutStrategy()
        href = strategy.get_service_desc_href(parent_dir="http://example.com")
        assert href == "http://example.com/api"

    def test_produces_layout_for_service_doc(self) -> None:
        strategy = APILayoutStrategy()
        href = strategy.get_service_doc_href(parent_dir="http://example.com")
        assert href == "http://example.com/api.html"



import json
import os
import shutil
import tempfile
from datetime import datetime, timezone
from typing import Any, cast

import jsonschema
import pytest
from requests_mock import Mocker

import pystac
import pystac.validation
from pystac.cache import CollectionCache
from pystac.serialization.common_properties import merge_common_properties
from pystac.utils import get_opt
from pystac.validation import GetSchemaError, JsonSchemaSTACValidator
from tests.utils import TestCases
from tests.utils.test_cases import ExampleInfo


class TestValidate:
    @pytest.mark.vcr()
    def test_validate_current_version(self) -> None:
        catalog = pystac.read_file(
            TestCases.get_path("data-files/catalogs/test-case-1/catalog.json")
        )
        catalog.validate()

        collection = pystac.read_file(
            TestCases.get_path(
                "data-files/catalogs/test-case-1//country-1/area-1-1/collection.json"
            )
        )
        collection.validate()

        item = pystac.read_file(TestCases.get_path("data-files/item/sample-item.json"))
        item.validate()

    @pytest.mark.vcr()
    @pytest.mark.parametrize("example", TestCases.get_examples_info())
    def test_validate_examples(self, example: ExampleInfo) -> None:
        stac_version = example.stac_version
        path = example.path
        valid = example.valid

        with open(path, encoding="utf-8") as f:
            stac_json = json.load(f)

        # Check if common properties need to be merged
        if stac_version < "1.0" and example.object_type == pystac.STACObjectType.ITEM:
            collection_cache = CollectionCache()
            merge_common_properties(stac_json, collection_cache, path)

        if valid:
            pystac.validation.validate_dict(stac_json)
        else:
            with pytest.raises(pystac.STACValidationError):
                try:
                    pystac.validation.validate_dict(stac_json)
                except pystac.STACValidationError as e:
                    assert isinstance(e.source, list)
                    assert isinstance(e.source[0], jsonschema.ValidationError)
                    raise e

    @pytest.mark.vcr()
    def test_validate_error_contains_href(self) -> None:
        # Test that the exception message contains the HREF of the object if available.
        cat = TestCases.case_1()
        item = next(cat.get_items("area-1-1-labels", recursive=True))
        assert item.get_self_href() is not None

        item.geometry = {"type": "INVALID"}

        with pytest.raises(pystac.STACValidationError):
            try:
                item.validate()
            except pystac.STACValidationError as e:
                assert get_opt(item.get_self_href()) in str(e)
                raise e

    @pytest.mark.vcr()
    def test_validate_all_deprecated_dict_arg(self) -> None:
        catalog = TestCases.case_1()

        with pytest.warns(DeprecationWarning, match="use validate_all_dict"):
            pystac.validation.validate_all(catalog.to_dict(), catalog.get_self_href())

    @pytest.mark.vcr()
    def test_validate_all_deprecated_dict_arg_missing_href(self) -> None:
        catalog = TestCases.case_1()

        with pytest.warns(DeprecationWarning, match="use validate_all_dict"):
            with pytest.raises(ValueError, match="href must be set"):
                pystac.validation.validate_all(catalog.to_dict())

    @pytest.mark.vcr()
    def test_validate_all_unexpected_href(self) -> None:
        catalog = TestCases.case_1()

        with pytest.raises(ValueError, match="href must be None"):
            pystac.validation.validate_all(catalog, catalog.get_self_href())

    @pytest.mark.vcr()
    def test_validate_all(self) -> None:
        catalog = TestCases.case_1()

        pystac.validation.validate_all(catalog)

    @pytest.mark.vcr()
    @pytest.mark.parametrize("test_case", TestCases.all_test_catalogs())
    def test_validate_all_dict(self, test_case: pystac.Catalog) -> None:
        catalog_href = test_case.get_self_href()
        if catalog_href is not None:
            stac_dict = pystac.StacIO.default().read_json(catalog_href)

            pystac.validation.validate_all_dict(stac_dict, catalog_href)

        # Modify a 0.8.1 collection in a catalog to be invalid with a
        # since-renamed extension and make sure it catches the validation error.

        with tempfile.TemporaryDirectory() as tmp_dir:
            dst_dir = os.path.join(tmp_dir, "catalog")
            # Copy test case 7 to the temporary directory
            catalog_href = get_opt(TestCases.case_7().get_self_href())
            shutil.copytree(os.path.dirname(catalog_href), dst_dir)

            new_cat_href = os.path.join(dst_dir, "catalog.json")

            # Make sure it's valid before modification
            pystac.validation.validate_all_dict(
                pystac.StacIO.default().read_json(new_cat_href), new_cat_href
            )

            # Modify a contained collection to add an extension for which the
            # collection is invalid.
            with open(
                os.path.join(dst_dir, "acc/collection.json"), encoding="utf-8"
            ) as f:
                col = json.load(f)
            col["stac_extensions"] = ["asset"]
            with open(
                os.path.join(dst_dir, "acc/collection.json"), "w", encoding="utf-8"
            ) as f:
                json.dump(col, f)

            stac_dict = pystac.StacIO.default().read_json(new_cat_href)

            with pytest.raises(pystac.STACValidationError):
                pystac.validation.validate_all_dict(stac_dict, new_cat_href)

    @pytest.mark.vcr()
    def test_validates_geojson_with_tuple_coordinates(self) -> None:
        """This unit tests guards against a bug where if a geometry
        dict has tuples instead of lists for the coordinate sequence,
        which can be produced by shapely, then the geometry still passes
        validation.
        """
        geom: dict[str, Any] = {
            "type": "Polygon",
            # Last , is required to ensure tuple creation.
            "coordinates": (
                (
                    (-115.305, 36.126),
                    (-115.305, 36.128),
                    (-115.307, 36.128),
                    (-115.307, 36.126),
                    (-115.305, 36.126),
                ),
            ),
        }

        item = pystac.Item(
            id="test-item",
            geometry=geom,
            bbox=[-115.308, 36.126, -115.305, 36.129],
            datetime=datetime.now(timezone.utc),
            properties={},
        )

        # Should not raise.
        item.validate()

    @pytest.mark.vcr()
    def test_validate_custom_validator(self) -> None:
        """This test verifies the use of a custom validator class passed as
        input to :meth:`~pystac.stac_object.STACObject.validate` and every
        underlying function. This validator is effective only for the call
        for which it was provided, contrary to
        :class:`~pystac.validation.RegisteredValidator`
        that persists it globally until reset.
        """
        custom_extension_uri = (
            "https://stac-extensions.github.io/custom-extension/v1.0.0/schema.json"
        )
        custom_extension_schema = {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "$id": f"{custom_extension_uri}#",
            "type": "object",
            "properties": {
                "properties": {
                    "type": "object",
                    "required": ["custom-extension:test"],
                    "properties": {"custom-extension:test": {"type": "integer"}},
                }
            },
        }
        item = cast(
            pystac.Item,
            pystac.read_file(TestCases.get_path("data-files/item/sample-item.json")),
        )
        item.stac_extensions.append(custom_extension_uri)
        item.properties["custom-extension:test"] = 123

        with pytest.raises(pystac.validation.GetSchemaError):
            item.validate()  # default validator does not know the extension

        class CustomValidator(JsonSchemaSTACValidator):
            def _get_schema(self, schema_uri: str) -> dict[str, Any]:
                if schema_uri == custom_extension_uri:
                    return custom_extension_schema
                return super()._get_schema(schema_uri)

        # validate that the custom schema is found with the custom validator
        custom_validator = CustomValidator()
        item.validate(validator=custom_validator)

        # make sure validation is effective
        item.properties["custom-extension:test"] = "bad-value"
        with pytest.raises(pystac.errors.STACValidationError):
            item.validate(validator=custom_validator)

        # verify that the custom validator is not persisted
        with pytest.raises(pystac.validation.GetSchemaError):
            item.validate()


@pytest.mark.block_network
def test_catalog_latest_version_uses_local(catalog: pystac.Catalog) -> None:
    assert catalog.validate()


@pytest.mark.block_network
def test_collection_latest_version_uses_local(collection: pystac.Collection) -> None:
    assert collection.validate()


@pytest.mark.block_network
def test_item_latest_version_uses_local(item: pystac.Item) -> None:
    assert item.validate()


def test_404_schema_url(requests_mock: Mocker, item: pystac.Item) -> None:
    requests_mock.get(
        "http://pystac-extensions.test/a-fake-schema.json", status_code=404
    )
    item.stac_extensions = ["http://pystac-extensions.test/a-fake-schema.json"]
    with pytest.raises(
        GetSchemaError, match="http://pystac-extensions.test/a-fake.schema.json"
    ):
        item.validate()



import json
import os
from datetime import datetime
from pathlib import Path

import pytest

import pystac
from tests.conftest import get_data_file
from tests.utils import ARBITRARY_BBOX, ARBITRARY_EXTENT, ARBITRARY_GEOM


def check_link(link: pystac.Link | None) -> None:
    assert link is not None
    href = link.get_target_str()
    assert href is not None
    assert "\\" not in href


def test_create_item_from_absolute_posix_href() -> None:
    absolute_href = get_data_file("item/sample-item.json")
    absolute_posix_href = absolute_href.replace("\\", "/")
    pystac.Item.from_file(absolute_posix_href)
    pystac.read_file(absolute_posix_href)


def test_create_item_from_relative_posix_href() -> None:
    absolute_href = get_data_file("item/sample-item.json")
    relative_href = os.path.relpath(absolute_href)
    relative_posix_href = relative_href.replace("\\", "/")
    pystac.Item.from_file(relative_posix_href)
    pystac.read_file(relative_posix_href)


def test_create_item_containing_posix_hrefs(tmp_path: Path) -> None:
    collection = pystac.Collection(
        "test-collection", "A test collection", ARBITRARY_EXTENT
    )
    item = pystac.Item("test-item", ARBITRARY_GEOM, ARBITRARY_BBOX, datetime.now(), {})
    collection.add_item(item)
    collection.normalize_and_save(
        str(tmp_path), catalog_type=pystac.CatalogType.ABSOLUTE_PUBLISHED
    )

    item_href = str(tmp_path / "test-item/test-item.json")
    item_dict = pystac.Item.from_file(item_href).to_dict()
    for link in item_dict["links"]:
        link["href"] = str(link["href"]).replace("\\\\", "/").replace("\\", "/")
    with open(item_href, "w") as f:
        json.dump(item_dict, f)

    collection_href = str(tmp_path / "collection.json")
    collection_dict = pystac.Collection.from_file(collection_href).to_dict()
    for link in collection_dict["links"]:
        link["href"] = str(link["href"]).replace("\\\\", "/").replace("\\", "/")
    with open(collection_href, "w") as f:
        json.dump(collection_dict, f)

    item_href = item_href.replace("\\", "/")
    pystac.Item.from_file(item_href)
    pystac.read_file(item_href)

    collection_href = collection_href.replace("\\", "/")
    pystac.Collection.from_file(collection_href)
    pystac.read_file(collection_href)


@pytest.mark.skipif(os.name != "nt", reason="windows only test")
def test_posix_self_link_from_absolute_href(tmp_path: Path) -> None:
    # Check that we convert to a windows style (\\) absolute href to posix style
    # in the self link
    absolute_windows_href = get_data_file("item/sample-item.json")
    assert "\\" in absolute_windows_href
    item = pystac.Item.from_file(absolute_windows_href)
    check_link(item.get_single_link(rel="self"))
    item2 = pystac.read_file(absolute_windows_href)
    check_link(item2.get_single_link(rel="self"))

    absolute_windows_href = get_data_file("collections/multi-extent.json")
    assert "\\" in absolute_windows_href
    collection = pystac.Collection.from_file(absolute_windows_href)
    check_link(collection.get_single_link(rel="self"))
    collection2 = pystac.read_file(absolute_windows_href)
    check_link(collection2.get_single_link(rel="self"))

    absolute_windows_href = get_data_file("catalogs/test-case-1/catalog.json")
    assert "\\" in absolute_windows_href
    catalog = pystac.Catalog.from_file(absolute_windows_href)
    check_link(catalog.get_single_link(rel="self"))
    catalog2 = pystac.read_file(absolute_windows_href)
    check_link(catalog2.get_single_link(rel="self"))

    # check that we retain a posix style (/) absolute href in the self link
    absolute_windows_href = get_data_file("item/sample-item.json")
    assert "\\" in absolute_windows_href
    absolute_posix_href = absolute_windows_href.replace("\\", "/")
    item = pystac.Item.from_file(absolute_posix_href)
    check_link(item.get_single_link(rel="self"))
    item2 = pystac.read_file(absolute_posix_href)
    check_link(item2.get_single_link(rel="self"))

    absolute_windows_href = get_data_file("collections/multi-extent.json")
    assert "\\" in absolute_windows_href
    absolute_posix_href = absolute_windows_href.replace("\\", "/")
    collection = pystac.Collection.from_file(absolute_posix_href)
    check_link(collection.get_single_link(rel="self"))
    collection2 = pystac.read_file(absolute_posix_href)
    check_link(collection2.get_single_link(rel="self"))

    absolute_windows_href = get_data_file("collections/multi-extent.json")
    assert "\\" in absolute_windows_href
    absolute_posix_href = absolute_windows_href.replace("\\", "/")
    catalog = pystac.Collection.from_file(absolute_posix_href)
    check_link(catalog.get_single_link(rel="self"))
    catalog2 = pystac.read_file(absolute_posix_href)
    check_link(catalog2.get_single_link(rel="self"))


@pytest.mark.skipif(os.name != "nt", reason="windows only test")
def test_posix_self_link_from_relative_href(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    # check that we convert a windows style (\\) relative href to posix style
    # in the self link (which is always converted to an absolute href)
    test_item = pystac.Item(
        "test-item", ARBITRARY_GEOM, ARBITRARY_BBOX, datetime.now(), {}
    )
    (tmp_path / "subdirectory").mkdir()
    test_item_href = str(tmp_path / "subdirectory" / "test-item.json")
    assert "\\" in test_item_href
    test_item.save_object(include_self_link=False, dest_href=test_item_href)

    monkeypatch.chdir(tmp_path)
    item = pystac.Item.from_file("subdirectory\\test-item.json")
    check_link(item.get_single_link(rel="self"))
    item2 = pystac.read_file("subdirectory\\test-item.json")
    check_link(item2.get_single_link(rel="self"))


def test_all_generated_links_have_posix_hrefs(tmp_path: Path) -> None:
    asset_href = Path(tmp_path / "test-asset-file.txt")
    asset_href.touch()
    asset = pystac.Asset(str(asset_href), media_type=pystac.MediaType.TEXT)

    item = pystac.Item("test-item", ARBITRARY_GEOM, ARBITRARY_BBOX, datetime.now(), {})
    item.add_asset("test-asset", asset)

    collection = pystac.Collection(
        "test-collection", "A test collection", ARBITRARY_EXTENT
    )
    collection.add_item(item)
    collection.normalize_hrefs(str(tmp_path))

    for link in collection.links:
        check_link(link)
    for link in item.links:
        check_link(link)


@pytest.mark.skipif(os.name != "nt", reason="windows only test")
def test_all_existing_links_converted_to_posix_hrefs() -> None:
    href = get_data_file("windows_hrefs/test-collection/test-item/test-item.json")
    item = pystac.Item.from_file(href)
    for link in item.links:
        check_link(link)
    with open(href) as f:
        item_dict = json.load(f)
    item2 = pystac.Item.from_dict(item_dict)
    for link in item2.links:
        check_link(link)

    href = get_data_file("windows_hrefs/test-collection/collection.json")
    collection = pystac.Collection.from_file(href)
    for link in collection.links:
        check_link(link)
    with open(href) as f:
        collection_dict = json.load(f)
    collection2 = pystac.Collection.from_dict(collection_dict)
    for link in collection2.links:
        check_link(link)

    href = get_data_file("windows_hrefs/catalog.json")
    catalog = pystac.Catalog.from_file(href)
    for link in catalog.links:
        check_link(link)
    with open(href) as f:
        catalog_dict = json.load(f)
    catalog2 = pystac.Catalog.from_dict(catalog_dict)
    for link in catalog2.links:
        check_link(link)


@pytest.mark.skipif(os.name != "nt", reason="windows only test")
def test_existing_asset_hrefs_converted_to_posix() -> None:
    href = get_data_file("windows_hrefs/test-collection/test-item/test-item.json")
    item = pystac.Item.from_file(href)
    for asset in item.get_assets().values():
        assert "\\" not in asset.href



"""Implements the :stac-ext:`Xarray Assets Extension <xarray>`."""

from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Generic, Literal, TypeVar

import pystac
from pystac.extensions.base import ExtensionManagementMixin, PropertiesExtension
from pystac.extensions.hooks import ExtensionHooks

#: Generalized version of :class:`~pystac.Collection`,
#: :class:`~pystac.Item`, or :class:`~pystac.Asset`
T = TypeVar("T", pystac.Collection, pystac.Item, pystac.Asset)

SCHEMA_URI = "https://stac-extensions.github.io/xarray-assets/v1.0.0/schema.json"

PREFIX: str = "xarray:"
OPEN_KWARGS_PROP = PREFIX + "open_kwargs"
STORAGE_OPTIONS_PROP = PREFIX + "storage_options"


class XarrayAssetsExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """An abstract class that can be used to extend the properties of a
    :class:`~pystac.Collection`, :class:`~pystac.Item`, or :class:`~pystac.Asset` with
    properties from the :stac-ext:`Xarray Assets Extension <xarray>`. This class is
    generic over the type of STAC Object to be extended (e.g. :class:`~pystac.Item`,
    :class:`~pystac.Asset`).

    To create a concrete instance of :class:`XarrayAssetsExtension`, use the
    :meth:`XarrayAssetsExtension.ext` method. For example:

    .. code-block:: python

        >>> item: pystac.Item = ...
        >>> xr_ext = XarrayAssetsExtension.ext(item)

    """

    name: Literal["xarray"] = "xarray"

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> XarrayAssetsExtension[T]:
        """Extend the given STAC Object with properties from the
        :stac-ext:`XarrayAssets Extension <xarray>`.

        This extension can be applied to instances of :class:`~pystac.Collection`,
        :class:`~pystac.Item` or :class:`~pystac.Asset`.

        Raises:
            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Collection):
            cls.ensure_has_extension(obj, add_if_missing)
            return CollectionXarrayAssetsExtension(obj)
        if isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return ItemXarrayAssetsExtension(obj)
        if isinstance(obj, pystac.Asset):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return AssetXarrayAssetsExtension(obj)
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))


class CollectionXarrayAssetsExtension(XarrayAssetsExtension[pystac.Collection]):
    """A concrete implementation of :class:`XarrayAssetsExtension` on a
    :class:`~pystac.Collection` that extends the properties of the Item to include
    properties defined in the :stac-ext:`XarrayAssets Extension <xarray>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`XarrayAssetsExtension.ext` on an :class:`~pystac.Collection` to extend it.
    """

    collection: pystac.Collection
    properties: dict[str, Any]

    def __init__(self, collection: pystac.Collection):
        self.collection = collection
        self.properties = collection.extra_fields

    def __repr__(self) -> str:
        return f"<CollectionXarrayAssetsExtension Item id={self.collection.id}>"


class ItemXarrayAssetsExtension(XarrayAssetsExtension[pystac.Item]):
    """A concrete implementation of :class:`XarrayAssetsExtension` on an
    :class:`~pystac.Item` that extends the properties of the Item to include properties
    defined in the :stac-ext:`XarrayAssets Extension <xarray>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`XarrayAssetsExtension.ext` on an :class:`~pystac.Item` to extend it.
    """

    item: pystac.Item
    properties: dict[str, Any]

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    def __repr__(self) -> str:
        return f"<ItemXarrayAssetsExtension Item id={self.item.id}>"


class AssetXarrayAssetsExtension(XarrayAssetsExtension[pystac.Asset]):
    """A concrete implementation of :class:`XarrayAssetsExtension` on an
    :class:`~pystac.Asset` that extends the Asset fields to include properties defined
    in the :stac-ext:`XarrayAssets Extension <xarray>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`XarrayAssetsExtension.ext` on an :class:`~pystac.Asset` to extend it.
    """

    asset: pystac.Asset
    properties: dict[str, Any]
    additional_read_properties: Iterable[dict[str, Any]] | None = None

    def __init__(self, asset: pystac.Asset):
        self.asset = asset
        self.properties = asset.extra_fields
        if asset.owner and isinstance(asset.owner, pystac.Item):
            self.additional_read_properties = [asset.owner.properties]

    @property
    def storage_options(self) -> dict[str, Any] | None:
        """Additional keywords for accessing the dataset from remote storage"""
        return self.properties.get(STORAGE_OPTIONS_PROP)

    @storage_options.setter
    def storage_options(self, v: dict[str, Any] | None) -> Any:
        if v is None:
            self.properties.pop(STORAGE_OPTIONS_PROP, None)
        else:
            self.properties[STORAGE_OPTIONS_PROP] = v

    @property
    def open_kwargs(self) -> dict[str, Any] | None:
        """Additional keywords for opening the dataset"""
        return self.properties.get(OPEN_KWARGS_PROP)

    @open_kwargs.setter
    def open_kwargs(self, v: dict[str, Any] | None) -> Any:
        if v is None:
            self.properties.pop(OPEN_KWARGS_PROP, None)
        else:
            self.properties[OPEN_KWARGS_PROP] = v

    def __repr__(self) -> str:
        return f"<AssetXarrayAssetsExtension Asset href={self.asset.href}>"


class XarrayAssetsExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids = {"xarray"}
    stac_object_types = {pystac.STACObjectType.COLLECTION, pystac.STACObjectType.ITEM}


XARRAY_ASSETS_EXTENSION_HOOKS: ExtensionHooks = XarrayAssetsExtensionHooks()



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, TypeVar

import pystac
from pystac.errors import STACError
from pystac.utils import (
    HREF as HREF,
)
from pystac.utils import (
    is_absolute_href,
    make_absolute_href,
    make_posix_style,
    make_relative_href,
)

if TYPE_CHECKING:
    from pystac.catalog import Catalog
    from pystac.collection import Collection
    from pystac.extensions.ext import LinkExt
    from pystac.item import Item
    from pystac.stac_object import STACObject

    PathLike = os.PathLike[str]

else:
    PathLike = os.PathLike

#: Generalized version of :class:`Link`
L = TypeVar("L", bound="Link")

#: Hierarchical links provide structure to STAC catalogs.
HIERARCHICAL_LINKS = [
    pystac.RelType.ROOT,
    pystac.RelType.CHILD,
    pystac.RelType.PARENT,
    pystac.RelType.COLLECTION,
    pystac.RelType.ITEM,
    pystac.RelType.ITEMS,
]


class Link(PathLike):
    """A link connects a :class:`~pystac.STACObject` to another entity.

    The target of a link can be either another STACObject, or
    an HREF. When serialized, links always refer to the HREF of the target.
    Links are lazily deserialized - this is, when you read in a link, it does
    not automatically load in the STACObject that is the target
    (if the link is pointing to a STACObject). When a user is crawling
    through a catalog or when additional metadata is required, PySTAC uses the
    :func:`~pystac.Link.resolve_stac_object` method
    to load in and deserialize STACObjects. This mechanism is used within
    the PySTAC codebase and normally does not need to be considered by the user -
    ideally the lazy deserialization of STACObjects is transparent to clients of PySTAC.

    Args:
        rel : The relation of the link (e.g. 'child', 'item'). Registered rel Types
            are preferred. See :class:`~pystac.RelType` for common media types.
        target : The target of the link. If the link is
            unresolved, or the link is to something that is not a STACObject,
            the target is an HREF. If resolved, the target is a STACObject.
        media_type : Optional description of the media type. Registered Media Types
            are preferred. See :class:`~pystac.MediaType` for common media types.
        title : Optional title for this link.
        extra_fields : Optional, additional fields for this link. This is used
            by extensions as a way to serialize and deserialize properties on link
            object JSON.
    """

    rel: str | pystac.RelType
    """The relation of the link (e.g. 'child', 'item'). Registered rel Types are
    preferred. See :class:`~pystac.RelType` for common media types."""

    media_type: str | pystac.MediaType | None
    """Optional description of the media type. Registered Media Types are preferred.
    See :class:`~pystac.MediaType` for common media types."""

    extra_fields: dict[str, Any]
    """Optional, additional fields for this link. This is used by extensions as a
    way to serialize and deserialize properties on link object JSON."""

    owner: STACObject | None
    """The owner of this link. The link will use its owner's root catalog
    :class:`~pystac.cache.ResolvedObjectCache` to resolve objects, and
    will create absolute HREFs from relative HREFs against the owner's self HREF."""

    _target_href: str | None
    _target_object: STACObject | None
    _title: str | None

    def __init__(
        self,
        rel: str | pystac.RelType,
        target: str | STACObject,
        media_type: str | pystac.MediaType | None = None,
        title: str | None = None,
        extra_fields: dict[str, Any] | None = None,
    ) -> None:
        self.rel = rel
        if isinstance(target, str):
            if rel == pystac.RelType.SELF:
                self._target_href = make_absolute_href(target)
            else:
                self._target_href = make_posix_style(target)
            self._target_object = None
        else:
            self._target_href = None
            self._target_object = target
        self.media_type = media_type
        self.title = title
        self.extra_fields = extra_fields or {}
        self.owner = None

    def set_owner(self, owner: STACObject | None) -> Link:
        """Sets the owner of this link.

        Args:
            owner: The owner of this link. Pass None to clear.
        """
        self.owner = owner
        return self

    @property
    def title(self) -> str | None:
        """Optional title for this link. If not provided during instantiation, this will
        attempt to get the title from the STAC object that the link references."""
        if self._title is not None:
            return self._title
        if self._target_object is not None and isinstance(
            self._target_object, pystac.Catalog
        ):
            return self._target_object.title
        return None

    @title.setter
    def title(self, v: str | None) -> None:
        self._title = v

    @property
    def href(self) -> str:
        """Returns the HREF for this link.

        If the href is None, this will throw an exception.
        Use get_href if there may not be an HREF.
        """
        result = self.get_href()
        if result is None:
            raise ValueError(f"{self} does not have an HREF set.")
        return result

    def get_href(self, transform_href: bool = True) -> str | None:
        """Gets the HREF for this link.

        Args:
            transform_href: If True, transform the HREF based on the type of
                catalog the owner belongs to (if any). I.e. if the link owner
                belongs to a root catalog that is RELATIVE_PUBLISHED or SELF_CONTAINED,
                the HREF will be transformed to be relative to the catalog root
                if this is a hierarchical link relation.

        Returns:
            str: Returns this link's HREF. If there is an owner of the link and
            the root catalog (if there is one) is of type RELATIVE_PUBLISHED,
            then the HREF returned will be relative.
            In all other cases, this method will return an absolute HREF.
        """
        # get the self href
        if self._target_object:
            href = self._target_object.get_self_href()
        else:
            href = self._target_href

        if (
            transform_href
            and href
            and is_absolute_href(href)
            and self.owner
            and self.owner.get_root()
        ):
            root = self.owner.get_root()
            rel_links = [
                *HIERARCHICAL_LINKS,
                *pystac.EXTENSION_HOOKS.get_extended_object_links(self.owner),
            ]
            # if a hierarchical link with an owner and root, and relative catalog
            if root and root.is_relative():
                if self.rel in rel_links or root.target_in_hierarchy(self.target):
                    owner_href = self.owner.get_self_href()
                    if owner_href is not None:
                        href = make_relative_href(href, owner_href)

        return href

    @property
    def absolute_href(self) -> str:
        """Returns the absolute HREF for this link.

        If the href is None, this will throw an exception.
        Use get_absolute_href if there may not be an HREF set.
        """
        result = self.get_absolute_href()
        if result is None:
            raise ValueError(f"{self} does not have an HREF set.")
        return result

    def get_absolute_href(self) -> str | None:
        """Gets the absolute href for this link, if possible.

        Returns:
            str: Returns this link's HREF. It attempts to derive an absolute HREF
            from this link; however, if the link is relative, has no owner,
            and has an unresolved target, this will return a relative HREF.
        """
        if self._target_object:
            href = self._target_object.get_self_href()
        else:
            href = self._target_href

        if href is not None and self.owner is not None:
            href = make_absolute_href(href, self.owner.get_self_href())

        return href

    @property
    def target(self) -> str | STACObject:
        """The target of the link. If the link is unresolved, or the link is to
        something that is not a STACObject, the target is an HREF. If resolved, the
        target is a STACObject."""
        if self._target_object:
            return self._target_object
        elif self._target_href:
            return self._target_href
        else:
            raise ValueError("No target defined for link.")

    @target.setter
    def target(self, target: str | STACObject) -> None:
        """Sets this link's target to a string or a STAC object."""
        if isinstance(target, str):
            self._target_href = target
            self._target_object = None
        else:
            self._target_href = None
            self._target_object = target

    def get_target_str(self) -> str | None:
        """Returns this link's target as a string.

        If a string href was provided, returns that. If not, tries to resolve
        the self link of the target object.
        """
        if self._target_href:
            return self._target_href
        elif self._target_object:
            return self._target_object.get_self_href()
        else:
            return None

    def has_target_href(self) -> bool:
        """Returns true if this link has a string href in its target information."""
        return self._target_href is not None

    def __fspath__(self) -> str:
        return self.absolute_href

    def __repr__(self) -> str:
        return f"<Link rel={self.rel} target={self.target}>"

    def _repr_html_(self) -> str:
        from html import escape

        from pystac.html.jinja_env import get_jinja_env

        jinja_env = get_jinja_env()
        if jinja_env:
            template = jinja_env.get_template("JSON.jinja2")
            return str(
                template.render(
                    dict=self.to_dict(transform_href=False), plain=escape(repr(self))
                )
            )
        else:
            return escape(repr(self))

    def resolve_stac_object(self, root: Catalog | None = None) -> Link:
        """Resolves a STAC object from the HREF of this link, if the link is not
        already resolved.

        Args:
            root : Optional root of the catalog for this link.
                If provided, the root's resolved object cache is used to search for
                previously resolved instances of the STAC object.
        """
        if self._target_object:
            pass
        elif self._target_href:
            target_href = self._target_href

            # If it's a relative link, base it off the parent.
            if not is_absolute_href(target_href):
                if self.owner is None:
                    raise pystac.STACError(
                        "Relative path {} encountered "
                        "without owner or start_href.".format(target_href)
                    )
                start_href = self.owner.get_self_href()

                if start_href is None:
                    raise pystac.STACError(
                        "Relative path {} encountered "
                        'without owner "self" link set.'.format(target_href)
                    )

                target_href = make_absolute_href(target_href, start_href)
            obj = None

            stac_io: pystac.StacIO | None = None

            if root is not None:
                obj = root._resolved_objects.get_by_href(target_href)
                stac_io = root._stac_io

            if obj is None:
                if stac_io is None:
                    if self.owner is not None:
                        if isinstance(self.owner, pystac.Catalog):
                            stac_io = self.owner._stac_io
                        elif self.rel != pystac.RelType.ROOT:
                            owner_root = self.owner.get_root()
                            if owner_root is not None:
                                stac_io = owner_root._stac_io
                    if stac_io is None:
                        stac_io = pystac.StacIO.default()
                try:
                    obj = stac_io.read_stac_object(target_href, root=root)
                except Exception as e:
                    raise STACError(
                        f"HREF: '{target_href}' does not resolve to a STAC object"
                    ) from e
                obj.set_self_href(target_href)
                if root is not None:
                    obj = root._resolved_objects.get_or_cache(obj)
                    obj.set_root(root)
            self._target_object = obj
        else:
            raise ValueError("Cannot resolve STAC object without a target")

        if (
            self.owner
            and self.rel in [pystac.RelType.CHILD, pystac.RelType.ITEM]
            and isinstance(self.owner, pystac.Catalog)
        ):
            assert self._target_object
            # Do nothing if the object wants to keep its parent
            # https://github.com/stac-utils/pystac/issues/1116
            if self._target_object._allow_parent_to_override_href:
                self._target_object.set_parent(self.owner)

        return self

    def is_resolved(self) -> bool:
        """Determines if the link's target is a resolved STACObject.

        Returns:
            bool: True if this link is resolved.
        """
        return self._target_object is not None

    def is_hierarchical(self) -> bool:
        """Returns true if this link's rel type is hierarchical.

        Hierarchical links are used to build relationships in STAC, e.g.
        "parent", "child", "item", etc. For a complete list of hierarchical
        relation types, see :py:const:`~pystac.link.HIERARCHICAL_LINKS`.

        Returns:
            bool: True if the link's rel type is hierarchical.
        """
        return self.rel in HIERARCHICAL_LINKS

    def to_dict(self, transform_href: bool = True) -> dict[str, Any]:
        """Returns this link as a dictionary.

        Args:
            transform_href : If ``True``, transform the HREF based on the type of
                catalog the owner belongs to (if any). I.e. if the link owner
                belongs to a root catalog that is RELATIVE_PUBLISHED or SELF_CONTAINED,
                the HREF will be transformed to be relative to the catalog root
                if this is a hierarchical link relation.
        Returns:
            dict : A serialization of the Link.
        """

        d: dict[str, Any] = {
            "rel": str(self.rel),
            "href": self.get_href(transform_href=transform_href),
        }

        if self.media_type is not None:
            d["type"] = str(self.media_type)

        if self.title is not None:
            d["title"] = self.title

        for k, v in self.extra_fields.items():
            d[k] = v

        return d

    def clone(self) -> Link:
        """Clones this link.

        This makes a copy of all link information, but does not clone a STACObject
        if one is the target of this link.

        Returns:
            Link: The cloned link.
        """
        cls = self.__class__
        return cls(
            rel=self.rel,
            target=self.target,
            media_type=self.media_type,
            title=self.title,
        )

    @classmethod
    def from_dict(cls: type[L], d: dict[str, Any]) -> L:
        """Deserializes a :class:`Link` from a dict.

        Args:
            d : The dict that represents the Link in JSON

        Returns:
            Link: Link instance constructed from the dict.
        """
        from copy import copy

        d = copy(d)
        rel = d.pop("rel")
        href = d.pop("href")
        media_type = d.pop("type", None)
        title = d.pop("title", None)

        extra_fields = None
        if any(d):
            extra_fields = d

        return cls(
            rel=rel,
            target=href,
            media_type=media_type,
            title=title,
            extra_fields=extra_fields,
        )

    @classmethod
    def root(cls: type[L], c: Catalog) -> L:
        """Creates a link to a root Catalog or Collection."""
        return cls(pystac.RelType.ROOT, c, media_type=pystac.MediaType.JSON)

    @classmethod
    def parent(cls: type[L], c: Catalog) -> L:
        """Creates a link to a parent Catalog or Collection."""
        return cls(pystac.RelType.PARENT, c, media_type=pystac.MediaType.JSON)

    @classmethod
    def collection(cls: type[L], c: Collection) -> L:
        """Creates a link to a Collection."""
        return cls(pystac.RelType.COLLECTION, c, media_type=pystac.MediaType.JSON)

    @classmethod
    def self_href(cls: type[L], href: HREF) -> L:
        """Creates a self link to a file's location."""
        href_str = str(os.fspath(href))
        return cls(pystac.RelType.SELF, href_str, media_type=pystac.MediaType.JSON)

    @classmethod
    def child(cls: type[L], c: Catalog, title: str | None = None) -> L:
        """Creates a link to a child Catalog or Collection."""
        return cls(
            pystac.RelType.CHILD, c, title=title, media_type=pystac.MediaType.JSON
        )

    @classmethod
    def item(cls: type[L], item: Item, title: str | None = None) -> L:
        """Creates a link to an Item."""
        return cls(
            pystac.RelType.ITEM, item, title=title, media_type=pystac.MediaType.GEOJSON
        )

    @classmethod
    def derived_from(cls: type[L], item: Item | str, title: str | None = None) -> L:
        """Creates a link to a derived_from Item."""
        return cls(
            pystac.RelType.DERIVED_FROM,
            item,
            title=title,
            media_type=pystac.MediaType.JSON,
        )

    @classmethod
    def canonical(
        cls: type[L],
        item_or_collection: Item | Collection,
        title: str | None = None,
    ) -> L:
        """Creates a canonical link to an Item or Collection."""
        return cls(
            pystac.RelType.CANONICAL,
            item_or_collection,
            title=title,
            media_type=pystac.MediaType.JSON,
        )

    @property
    def ext(self) -> LinkExt:
        """Accessor for extension classes on this link

        Example::

            link.ext.file.size = 8675309
        """
        from pystac.extensions.ext import LinkExt

        return LinkExt(stac_object=self)



# isort: skip_file
"""
PySTAC is a library for working with SpatioTemporal Asset Catalogs (STACs)
"""

__all__ = [
    "__version__",
    "TemplateError",
    "STACError",
    "STACTypeError",
    "DuplicateObjectKeyError",
    "ExtensionAlreadyExistsError",
    "ExtensionNotImplemented",
    "ExtensionTypeError",
    "RequiredPropertyMissing",
    "STACValidationError",
    "DeprecatedWarning",
    "MediaType",
    "RelType",
    "StacIO",
    "STACObject",
    "STACObjectType",
    "Link",
    "HIERARCHICAL_LINKS",
    "Catalog",
    "CatalogType",
    "Collection",
    "Extent",
    "SpatialExtent",
    "TemporalExtent",
    "Summaries",
    "CommonMetadata",
    "RangeSummary",
    "Item",
    "Asset",
    "ItemAssetDefinition",
    "ItemCollection",
    "Provider",
    "ProviderRole",
    "read_file",
    "read_dict",
    "write_file",
    "get_stac_version",
    "set_stac_version",
]

import warnings
from typing import Any

from pystac.errors import (
    TemplateError,
    STACError,
    STACTypeError,
    DuplicateObjectKeyError,
    ExtensionAlreadyExistsError,
    ExtensionNotImplemented,
    ExtensionTypeError,
    RequiredPropertyMissing,
    STACValidationError,
    DeprecatedWarning,
)

from pystac.version import (
    __version__,
    get_stac_version,
    set_stac_version,
)
from pystac.media_type import MediaType
from pystac.rel_type import RelType
from pystac.stac_io import StacIO
from pystac.stac_object import STACObject, STACObjectType
from pystac.link import Link, HIERARCHICAL_LINKS
from pystac.catalog import Catalog, CatalogType
from pystac.collection import (
    Collection,
    Extent,
    SpatialExtent,
    TemporalExtent,
)
from pystac.common_metadata import CommonMetadata
from pystac.summaries import RangeSummary, Summaries
from pystac.asset import Asset
from pystac.item import Item
from pystac.item_assets import ItemAssetDefinition
from pystac.item_collection import ItemCollection
from pystac.provider import ProviderRole, Provider
from pystac.utils import HREF

import pystac.extensions.hooks
import pystac.extensions.classification
import pystac.extensions.datacube
import pystac.extensions.eo
import pystac.extensions.file
import pystac.extensions.grid
import pystac.extensions.item_assets

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    import pystac.extensions.label
import pystac.extensions.mgrs
import pystac.extensions.mlm
import pystac.extensions.pointcloud
import pystac.extensions.projection
import pystac.extensions.raster
import pystac.extensions.sar
import pystac.extensions.sat
import pystac.extensions.scientific
import pystac.extensions.storage
import pystac.extensions.table
import pystac.extensions.timestamps
import pystac.extensions.version
import pystac.extensions.view
import pystac.extensions.xarray_assets

EXTENSION_HOOKS = pystac.extensions.hooks.RegisteredExtensionHooks(
    [
        pystac.extensions.classification.CLASSIFICATION_EXTENSION_HOOKS,
        pystac.extensions.datacube.DATACUBE_EXTENSION_HOOKS,
        pystac.extensions.eo.EO_EXTENSION_HOOKS,
        pystac.extensions.file.FILE_EXTENSION_HOOKS,
        pystac.extensions.grid.GRID_EXTENSION_HOOKS,
        pystac.extensions.item_assets.ITEM_ASSETS_EXTENSION_HOOKS,
        pystac.extensions.label.LABEL_EXTENSION_HOOKS,
        pystac.extensions.mgrs.MGRS_EXTENSION_HOOKS,
        pystac.extensions.mlm.MLM_EXTENSION_HOOKS,
        pystac.extensions.pointcloud.POINTCLOUD_EXTENSION_HOOKS,
        pystac.extensions.projection.PROJECTION_EXTENSION_HOOKS,
        pystac.extensions.raster.RASTER_EXTENSION_HOOKS,
        pystac.extensions.sar.SAR_EXTENSION_HOOKS,
        pystac.extensions.sat.SAT_EXTENSION_HOOKS,
        pystac.extensions.scientific.SCIENTIFIC_EXTENSION_HOOKS,
        pystac.extensions.storage.STORAGE_EXTENSION_HOOKS,
        pystac.extensions.table.TABLE_EXTENSION_HOOKS,
        pystac.extensions.timestamps.TIMESTAMPS_EXTENSION_HOOKS,
        pystac.extensions.version.VERSION_EXTENSION_HOOKS,
        pystac.extensions.view.VIEW_EXTENSION_HOOKS,
        pystac.extensions.xarray_assets.XARRAY_ASSETS_EXTENSION_HOOKS,
    ]
)


def read_file(href: HREF, stac_io: StacIO | None = None) -> STACObject:
    """Reads a STAC object from a file.

    This method will return either a Catalog, a Collection, or an Item based on what
    the file contains.

    This is a convenience method for :meth:`StacIO.read_stac_object
    <pystac.StacIO.read_stac_object>`

    Args:
        href : The HREF to read the object from.
        stac_io: Optional :class:`~StacIO` instance to use for I/O operations. If not
            provided, will use :meth:`StacIO.default` to create an instance.

    Returns:
        The specific STACObject implementation class that is represented
        by the JSON read from the file located at HREF.

    Raises:
        STACTypeError : If the file at ``href`` does not represent a valid
            :class:`~pystac.STACObject`. Note that an :class:`~pystac.ItemCollection`
            is not a :class:`~pystac.STACObject` and must be read using
            :meth:`ItemCollection.from_file <pystac.ItemCollection.from_file>`
    """
    if stac_io is None:
        stac_io = StacIO.default()
    return stac_io.read_stac_object(href)


def write_file(
    obj: STACObject,
    include_self_link: bool = True,
    dest_href: HREF | None = None,
    stac_io: StacIO | None = None,
) -> None:
    """Writes a STACObject to a file.

    This will write only the Catalog, Collection or Item ``obj``. It will not attempt
    to write any other objects that are linked to ``obj``; if you'd like functionality
    to save off catalogs recursively see :meth:`Catalog.save <pystac.Catalog.save>`.

    This method will write the JSON of the object to the object's assigned "self" link
    or to the dest_href if provided. To set the self link, see
    :meth:`STACObject.set_self_href <pystac.STACObject.set_self_href>`.

    Convenience method for :meth:`STACObject.from_file <pystac.STACObject.from_file>`

    Args:
        obj : The STACObject to save.
        include_self_link : If ``True``, include the ``"self"`` link with this object.
            Otherwise, leave out the self link.
        dest_href : Optional HREF to save the file to. If ``None``, the object will be
            saved to the object's ``"self"`` href.
        stac_io: Optional :class:`~StacIO` instance to use for I/O operations. If not
            provided, will use :meth:`StacIO.default` to create an instance.
    """
    if stac_io is None:
        stac_io = StacIO.default()
    import os

    dest_href = None if dest_href is None else str(os.fspath(dest_href))
    obj.save_object(
        include_self_link=include_self_link, dest_href=dest_href, stac_io=stac_io
    )


def read_dict(
    d: dict[str, Any],
    href: str | None = None,
    root: Catalog | None = None,
    stac_io: StacIO | None = None,
) -> STACObject:
    """Reads a :class:`~STACObject` or :class:`~ItemCollection` from a JSON-like dict
    representing a serialized STAC object.

    This method will return either a :class:`~Catalog`, :class:`~Collection`,
    or :class`~Item` based on the contents of the dict.

    This is a convenience method for either
    :meth:`StacIO.stac_object_from_dict <pystac.StacIO.stac_object_from_dict>`.

    Args:
        d : The dict to parse.
        href : Optional href that is the file location of the object being
            parsed.
        root : Optional root of the catalog for this object.
            If provided, the root's resolved object cache can be used to search for
            previously resolved instances of the STAC object.
        stac_io: Optional :class:`~StacIO` instance to use for reading. If ``None``,
            the default instance will be used.

    Raises:
        STACTypeError : If the ``d`` dictionary does not represent a valid
            :class:`~pystac.STACObject`. Note that an :class:`~pystac.ItemCollection`
            is not a :class:`~pystac.STACObject` and must be read using
            :meth:`ItemCollection.from_dict <pystac.ItemCollection.from_dict>`
    """
    if stac_io is None:
        stac_io = StacIO.default()
    return stac_io.stac_object_from_dict(d, href, root)


def __getattr__(name: str) -> Any:
    if name == "validation":
        import warnings
        import pystac.validation

        warnings.warn(
            "pystac.validation will not be automatically imported to the package in "
            "pystac v2.0. Instead, import it directly: `import pystac.validation`",
            DeprecationWarning,
            stacklevel=2,
        )
        return pystac.validation
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



import json
import logging
from copy import deepcopy
from datetime import datetime
from typing import Any, cast

import pytest
from dateutil.parser import parse

import pystac
from pystac import Collection, Item
from pystac.extensions.classification import (
    CLASSES_PROP,
    DEFAULT_VERSION,
    SCHEMA_URI_PATTERN,
    SUPPORTED_VERSIONS,
    Bitfield,
    Classification,
    ClassificationExtension,
)
from pystac.extensions.raster import RasterBand, RasterExtension
from tests.utils import TestCases

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger()

LANDSAT_EXAMPLE_URI = TestCases.get_path(
    "data-files/classification/classification_landsat_example.json"
)
CLASSIFICATION_COLLECTION_RASTER_URI = TestCases.get_path(
    "data-files/classification/collection-item-assets-raster-bands.json"
)
PLAIN_ITEM = TestCases.get_path("data-files/item/sample-item.json")


@pytest.fixture
def item_dict() -> dict[str, Any]:
    with open(LANDSAT_EXAMPLE_URI) as f:
        return cast(dict[str, Any], json.load(f))


@pytest.fixture
def landsat_item() -> Item:
    return Item.from_file(LANDSAT_EXAMPLE_URI)


@pytest.fixture
def plain_item() -> Item:
    return Item.from_file(PLAIN_ITEM)


@pytest.fixture
def classification_collection() -> Collection:
    return Collection.from_file(CLASSIFICATION_COLLECTION_RASTER_URI)


def test_stac_extensions(landsat_item: Item) -> None:
    assert ClassificationExtension.has_extension(landsat_item)


def test_classification_object() -> None:
    c = Classification.create(
        name="dummy",
        description="empty class",
        value=0,
        title="dummy title",
        color_hint="FF00AB",
        nodata=True,
        percentage=20.3,
        count=2,
    )
    assert c.name == "dummy"
    assert c.description == "empty class"
    assert c.color_hint == "FF00AB"
    assert c.value == 0
    assert c.title == "dummy title"
    assert c.nodata is True
    assert c.percentage == 20.3
    assert c.count == 2

    assert Classification(c.to_dict()) == c
    with pytest.raises(NotImplementedError):
        c == "blah"


def test_bitfield_object() -> None:
    b = Bitfield.create(
        offset=0,
        length=1,
        classes=[
            Classification.create(name="no", value=0),
            Classification.create(name="yes", value=1),
        ],
        roles=["data"],
        description="dummy description",
        name="dummy",
    )
    assert b.offset == 0
    assert b.length == 1
    assert len(b.classes) == 2
    assert b.roles == ["data"]
    assert b.description == "dummy description"
    assert b.name == "dummy"


def test_get_schema_uri(landsat_item: Item) -> None:
    with pytest.raises(DeprecationWarning):
        assert any(
            [
                uri in landsat_item.stac_extensions
                for uri in ClassificationExtension.get_schema_uris()
            ]
        )


def test_ext_raises_if_item_does_not_conform(plain_item: Item) -> None:
    with pytest.raises(pystac.errors.ExtensionNotImplemented):
        ClassificationExtension.ext(plain_item)


def test_ext_raises_on_collection(classification_collection: Collection) -> None:
    with pytest.raises(
        pystac.errors.ExtensionTypeError,
        match="ClassificationExtension does not apply to type 'Collection'",
    ) as e:
        ClassificationExtension.ext(classification_collection)  # type:ignore
    assert "Hint" in str(e.value)


@pytest.mark.vcr()
def test_apply_bitfields(plain_item: Item) -> None:
    ClassificationExtension.add_to(plain_item)
    ClassificationExtension.ext(plain_item).apply(
        bitfields=[
            Bitfield.create(
                offset=0,
                length=1,
                classes=[
                    Classification.create(name="no", value=0),
                    Classification.create(name="yes", value=1),
                ],
            )
        ]
    )

    plain_item.validate()
    assert (
        ClassificationExtension.ext(plain_item).bitfields is not None
        and len(cast(list[Bitfield], ClassificationExtension.ext(plain_item).bitfields))
        == 1
    )
    assert (
        cast(list[Bitfield], ClassificationExtension.ext(plain_item).bitfields)[
            0
        ].offset
        == 0
    )
    assert (
        cast(list[Bitfield], ClassificationExtension.ext(plain_item).bitfields)[
            0
        ].length
        == 1
    )
    assert (
        ClassificationExtension.ext(plain_item).bitfields is not None
        and len(
            cast(list[Bitfield], ClassificationExtension.ext(plain_item).bitfields)[
                0
            ].classes
        )
        == 2
    )


@pytest.mark.vcr()
def test_apply_classes(plain_item: Item) -> None:
    ClassificationExtension.add_to(plain_item)
    ClassificationExtension.ext(plain_item).apply(
        classes=[
            Classification.create(name="no", value=0),
            Classification.create(name="yes", value=1),
        ]
    )
    plain_item.validate()
    assert (
        ClassificationExtension.ext(plain_item).classes is not None
        and len(
            cast(list[Classification], ClassificationExtension.ext(plain_item).classes)
        )
        == 2
    )


def test_create_classes(plain_item: Item) -> None:
    ClassificationExtension.add_to(plain_item)
    ext = ClassificationExtension.ext(plain_item)
    ext.apply(
        bitfields=[
            Bitfield.create(
                offset=0,
                length=1,
                classes=[
                    Classification.create(name="no", value=0),
                    Classification.create(name="yes", value=1),
                ],
            )
        ]
    )
    ext.classes = [
        Classification.create(name="no", value=0),
        Classification.create(name="yes", value=1),
    ]
    assert ext.bitfields is None
    ext.bitfields = [
        Bitfield.create(
            offset=0,
            length=1,
            classes=[
                Classification.create(name="no", value=0),
                Classification.create(name="yes", value=1),
            ],
        )
    ]
    assert ext.classes is None


def test_create() -> None:
    field = Bitfield.create(
        name="cloud_confidence",
        description="Cloud confidence levels",
        offset=8,
        length=2,
        classes=[
            Classification.create(
                name="not_set", description="No confidence level set", value=0
            ),
            Classification.create(
                name="low", description="Low confidence cloud", value=1
            ),
            Classification.create(
                name="medium", description="Medium confidence cloud", value=2
            ),
            Classification.create(
                name="high", description="High confidence cloud", value=3
            ),
        ],
    )

    logger.info(field)


def test_color_hint_formatting() -> None:
    with pytest.raises(Exception):
        Classification.create(value=0, name="water", color_hint="#0000ff")
    Classification.create(value=0, name="water", color_hint="0000FF")


def test_to_from_dict(item_dict: dict[str, Any]) -> None:
    def _parse_times(a_dict: dict[str, Any]) -> None:
        for k, v in a_dict.items():
            if isinstance(v, dict):
                _parse_times(v)
            elif isinstance(v, (tuple, list, set)):
                for vv in v:
                    if isinstance(vv, dict):
                        _parse_times(vv)
            else:
                if k == "datetime":
                    if not isinstance(v, datetime):
                        a_dict[k] = parse(v)
                        a_dict[k] = a_dict[k].replace(microsecond=0)

    d1 = deepcopy(item_dict)
    d2 = Item.from_dict(item_dict, migrate=False).to_dict()
    _parse_times(d1)
    _parse_times(d2)
    assert d1 == d2, f"Mismatch between dictionaries: \n{d1}\n{d2}"


def test_add_to(plain_item: Item) -> None:
    assert ClassificationExtension.get_schema_uri() not in plain_item.stac_extensions

    # Check that the URI gets added to stac_extensions
    ClassificationExtension.add_to(plain_item)
    assert ClassificationExtension.get_schema_uri() in plain_item.stac_extensions

    # Check that the URI only gets added once, regardless of how many times add_to
    # is called.
    ClassificationExtension.add_to(plain_item)
    ClassificationExtension.add_to(plain_item)

    classification_uris = [
        uri
        for uri in plain_item.stac_extensions
        if uri == ClassificationExtension.get_schema_uri()
    ]
    assert len(classification_uris) == 1


@pytest.mark.vcr()
def test_validate_classification(landsat_item: Item) -> None:
    landsat_item.validate()


def test_add_item_classes(plain_item: Item) -> None:
    item_ext = ClassificationExtension.ext(plain_item, add_if_missing=True)
    item_ext.__repr__()
    assert item_ext.classes is None
    item_ext.classes = [Classification.create(name="dummy", value=0)]
    assert item_ext.properties[CLASSES_PROP] == [{"value": 0, "name": "dummy"}]


def test_add_asset_classes(plain_item: Item) -> None:
    ClassificationExtension.ext(plain_item, add_if_missing=True)
    asset = plain_item.assets["analytic"]
    assert CLASSES_PROP not in asset.extra_fields.keys()
    asset_ext = ClassificationExtension.ext(asset)
    asset_ext.__repr__()
    asset_ext.classes = [Classification.create(value=0, name="dummy")]
    assert CLASSES_PROP in asset.extra_fields.keys()
    assert asset.extra_fields[CLASSES_PROP] == [{"value": 0, "name": "dummy"}]


def test_item_asset_raster_classes(classification_collection: Collection) -> None:
    assert classification_collection.item_assets
    item_asset = classification_collection.item_assets["cloud-mask-raster"]
    raster_bands = cast(list[RasterBand], RasterExtension.ext(item_asset).bands)
    raster_bands_ext = ClassificationExtension.ext(raster_bands[0])
    raster_bands_ext.__repr__()
    assert raster_bands_ext.classes is not None


def test_item_assets_extension(classification_collection: Collection) -> None:
    assert classification_collection.item_assets
    item_asset = classification_collection.item_assets["cloud-mask-raster"]
    ext = ClassificationExtension.ext(item_asset)
    ext.__repr__()
    assert (
        ClassificationExtension.get_schema_uri()
        in classification_collection.stac_extensions
    )
    assert classification_collection.item_assets["cloud-mask-raster"].ext.has(
        "classification"
    )


def test_older_extension_version(landsat_item: Item) -> None:
    OLD_VERSION = list(set(SUPPORTED_VERSIONS) - {DEFAULT_VERSION})[0]
    new = SCHEMA_URI_PATTERN.format(version=DEFAULT_VERSION)
    old = SCHEMA_URI_PATTERN.format(version=OLD_VERSION)

    stac_extensions = set(landsat_item.stac_extensions)
    stac_extensions.remove(new)
    stac_extensions.add(old)
    item_as_dict = landsat_item.to_dict(include_self_link=False, transform_hrefs=False)
    item_as_dict["stac_extensions"] = list(stac_extensions)
    item = Item.from_dict(item_as_dict, migrate=False)
    assert ClassificationExtension.has_extension(item)
    assert old in item.stac_extensions

    migrated_item = pystac.Item.from_dict(item_as_dict, migrate=True)
    assert ClassificationExtension.has_extension(migrated_item)
    assert new in migrated_item.stac_extensions


def test_migrate_color_hint(plain_item: Item) -> None:
    item_dict = plain_item.to_dict(include_self_link=False, transform_hrefs=False)
    item_dict["stac_extensions"] = [
        "https://stac-extensions.github.io/classification/v1.0.0/schema.json"
    ]
    item_dict["assets"]["analytic"]["classification:classes"] = [
        {"value": 42, "color-hint": "#ffffff"}
    ]
    item = Item.from_dict(item_dict, migrate=True)
    classification = ClassificationExtension.ext(item.assets["analytic"])
    assert classification.classes
    assert classification.classes[0].color_hint == "#ffffff"



from __future__ import annotations

from typing import TYPE_CHECKING, Any
from unittest.mock import Mock

import pystac
from pystac.stac_io import DefaultStacIO, StacIO

if TYPE_CHECKING:
    from pystac.utils import HREF


class MockStacIO(pystac.StacIO):
    """Creates a mock that records StacIO calls for testing and allows
    clients to replace StacIO functionality, all within a context scope.

    Args:
        wrapped_stac_io: The StacIO that will be used to perform the calls.
            Defaults to an instance of DefaultStacIO.
    """

    mock: Mock
    wrapped_stac_io: StacIO

    def __init__(self, wrapped_stac_io: StacIO | None = None) -> None:
        self.mock = Mock()
        if wrapped_stac_io is None:
            self.wrapped_stac_io = DefaultStacIO()
        else:
            self.wrapped_stac_io = wrapped_stac_io

    def read_text(self, source: HREF, *args: Any, **kwargs: Any) -> str:
        self.mock.read_text(source)
        return self.wrapped_stac_io.read_text(source)

    def write_text(
        self,
        dest: HREF,
        txt: str,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.mock.write_text(dest, txt)
        self.wrapped_stac_io.write_text(dest, txt)


class MockDefaultStacIO:
    """Context manager for mocking StacIO."""

    def __enter__(self) -> MockStacIO:
        mock = MockStacIO()
        pystac.StacIO.set_default(lambda: mock)
        return mock

    def __exit__(self, *args: Any) -> None:
        pystac.StacIO.set_default(DefaultStacIO)



__version__ = "1.14.3"  # x-release-please-version
"""Library version"""


class STACVersion:
    DEFAULT_STAC_VERSION = "1.1.0"
    """Latest STAC version supported by PySTAC"""

    DEFAULT_STAC_API_VERSION = "1.0.0"
    """Latest STAC API version supported by PySTAC"""

    # Version that holds a user-set STAC version to use.
    _override_version: str | None = None

    OVERRIDE_VERSION_ENV_VAR = "PYSTAC_STAC_VERSION_OVERRIDE"

    @classmethod
    def get_stac_version(cls) -> str:
        if cls._override_version is not None:
            return cls._override_version

        import os

        env_version = os.environ.get(cls.OVERRIDE_VERSION_ENV_VAR)
        if env_version is not None:
            return env_version

        return cls.DEFAULT_STAC_VERSION

    @classmethod
    def set_stac_version(cls, stac_version: str | None) -> None:
        cls._override_version = stac_version


def get_stac_version() -> str:
    """Returns the STAC version PySTAC writes as the "stac_version" property for
    any object it serializes into JSON.

    If a call to ``set_stac_version`` was made, this will return the value it was
    called with. Next it will check the environment for a PYSTAC_STAC_VERSION_OVERRIDE
    variable. Otherwise, it will return the latest STAC version that this version of
    PySTAC supports.

    Returns:
        str: The STAC Version PySTAC is set up to use.
    """
    return STACVersion.get_stac_version()


def set_stac_version(stac_version: str | None) -> None:
    """Sets the STAC version that PySTAC should use.

    This is the version that will be set as the "stac_version" property
    on any JSON STAC objects written by PySTAC. If set to None, the override version
    will be cleared if previously set and the default or an override taken from the
    environment will be used.

    You can also set the environment variable PYSTAC_STAC_VERSION_OVERRIDE to override
    the version.

    Args:
        stac_version : The STAC version to use instead of the latest STAC version
            that PySTAC supports (described in STACVersion.DEFAULT_STAC_VERSION).
            If None, clear to use the default for this version of PySTAC.

    Note:
        Setting the STAC version to something besides the default version will not
        affect the format of STAC read or written; it will only override the
        ``stac_version`` property of the objects being written. Setting this
        incorrectly can produce invalid STAC.
    """
    STACVersion.set_stac_version(stac_version)



import pystac
from pystac.validation.schema_uri_map import DefaultSchemaUriMap


def test_gets_schema_uri_for_old_version() -> None:
    d = DefaultSchemaUriMap()
    uri = d.get_object_schema_uri(pystac.STACObjectType.ITEM, "0.8.0")

    assert uri == (
        "https://raw.githubusercontent.com/radiantearth/stac-spec/v0.8.0/"
        "item-spec/json-schema/item.json"
    )



"""Tests for pystac.tests.extensions.datacube"""

import json
from typing import Any

import pytest

import pystac
import pystac.extensions.datacube as dc
from pystac import Asset, Item
from pystac.errors import ExtensionTypeError
from tests.conftest import get_data_file


@pytest.fixture
def ext_item_uri() -> str:
    return get_data_file("datacube/item.json")


@pytest.fixture
def ext_item(ext_item_uri: str) -> Item:
    return Item.from_file(ext_item_uri)


@pytest.fixture
def horizontal_dimension(ext_item: Item) -> dc.Dimension:
    return dc.DatacubeExtension.ext(ext_item).dimensions["x"]


@pytest.fixture
def temporal_dimension(ext_item: Item) -> dc.Dimension:
    return dc.DatacubeExtension.ext(ext_item).dimensions["time"]


@pytest.fixture
def additional_dimension(ext_item: Item) -> dc.Dimension:
    return dc.DatacubeExtension.ext(ext_item).dimensions["spectral"]


@pytest.mark.parametrize(
    "field,value",
    [
        ("axis", "y"),
        ("extent", [-102, -120]),
        ("reference_system", 100),
    ],
)
def test_horizontal_dimension_fields(
    horizontal_dimension: dc.Dimension, field: str, value: Any
) -> None:
    dim = horizontal_dimension
    original = dim.properties[field]
    assert original == getattr(dim, field)

    setattr(dim, field, value)
    new = dim.properties[field]

    assert new != original
    assert new == value


def test_horizontal_dimension_values(
    horizontal_dimension: dc.HorizontalSpatialDimension,
) -> None:
    dim = horizontal_dimension
    assert dim.values is None

    dim.values = [0, 180]
    assert dim.values == [0, 180]
    assert dim.properties["values"] == [0, 180]

    dim.values = None
    assert dim.values is None
    assert "values" not in dim.properties


def test_horizontal_dimension_step(
    horizontal_dimension: dc.HorizontalSpatialDimension,
) -> None:
    dim = horizontal_dimension
    assert dim.step is None

    dim.step = 3
    assert dim.step == 3
    assert dim.properties["step"] == 3

    dim.clear_step()
    assert dim.step is None
    assert "step" not in dim.properties


def test_temporal_dimension_type(temporal_dimension: dc.TemporalDimension) -> None:
    assert isinstance(temporal_dimension, dc.TemporalDimension)
    assert temporal_dimension.dim_type == "temporal"

    temporal_dimension.dim_type = "foo"
    assert temporal_dimension.dim_type == "foo"


def test_temporal_dimension_description(
    temporal_dimension: dc.TemporalDimension,
) -> None:
    assert temporal_dimension.description is None
    value = "Dimension in time"

    temporal_dimension.description = value
    assert temporal_dimension.properties["description"] == value

    temporal_dimension.description = None
    assert "description" not in temporal_dimension.properties


def test_vertical_dimension_extent_not_required() -> None:
    props: dict[str, list[float]] = {}
    dim = dc.VerticalSpatialDimension(props)
    assert dim.extent is None


def test_vertical_dimension_setting_extent_to_none_pops_it() -> None:
    props: dict[str, list[float]] = {"extent": [10, 100]}
    dim = dc.VerticalSpatialDimension(props)
    assert dim.extent == [10, 100]
    dim.extent = None
    assert props == {}


def test_stac_extensions(ext_item: Item) -> None:
    assert dc.DatacubeExtension.has_extension(ext_item)


def test_get_schema_uri(ext_item: Item) -> None:
    assert dc.DatacubeExtension.get_schema_uri() in ext_item.stac_extensions


def test_ext_raises_if_item_does_not_conform(item: Item) -> None:
    with pytest.raises(pystac.errors.ExtensionNotImplemented):
        dc.DatacubeExtension.ext(item)


def test_ext_raises_if_asset_does_not_conform(sample_item: Item) -> None:
    asset = sample_item.assets["analytic"]
    with pytest.raises(pystac.errors.ExtensionNotImplemented):
        dc.DatacubeExtension.ext(asset)


def test_ext_always_passes_if_asset_has_no_owner(sample_item: Item) -> None:
    asset = sample_item.assets["analytic"]

    ownerless_asset = pystac.Asset.from_dict(asset.to_dict())
    dc.DatacubeExtension.ext(ownerless_asset)


def test_ext_raises_if_passed_a_non_stac_object() -> None:
    with pytest.raises(
        ExtensionTypeError, match="DatacubeExtension does not apply to type 'object'"
    ):
        dc.DatacubeExtension.ext(object())  # type: ignore


def test_to_from_dict(ext_item_uri: str, ext_item: Item) -> None:
    with open(ext_item_uri) as f:
        d = json.load(f)
    actual = ext_item.to_dict(include_self_link=False)
    assert d.keys() == actual.keys()
    for k, v in d.items():
        if k not in ["links"]:
            assert actual[k] == v


def test_add_to_collection(collection: pystac.Collection) -> None:
    assert not dc.DatacubeExtension.has_extension(collection)
    dc.DatacubeExtension.add_to(collection)

    assert dc.DatacubeExtension.has_extension(collection)
    assert dc.DatacubeExtension.get_schema_uri() in collection.stac_extensions


def test_add_ext_to_collection(collection: pystac.Collection) -> None:
    assert not dc.DatacubeExtension.has_extension(collection)
    dc.DatacubeExtension.ext(collection, add_if_missing=True)

    assert dc.DatacubeExtension.has_extension(collection)
    assert dc.DatacubeExtension.get_schema_uri() in collection.stac_extensions


def test_add_to_item(item: Item) -> None:
    assert not dc.DatacubeExtension.has_extension(item)
    dc.DatacubeExtension.add_to(item)

    assert dc.DatacubeExtension.has_extension(item)
    assert dc.DatacubeExtension.get_schema_uri() in item.stac_extensions


def test_add_ext_to_item(item: Item) -> None:
    assert not dc.DatacubeExtension.has_extension(item)
    dc.DatacubeExtension.ext(item, add_if_missing=True)

    assert dc.DatacubeExtension.has_extension(item)
    assert dc.DatacubeExtension.get_schema_uri() in item.stac_extensions


def test_add_ext_to_asset(sample_item: Item) -> None:
    assert not dc.DatacubeExtension.has_extension(sample_item)
    asset = sample_item.assets["analytic"]
    dc.DatacubeExtension.ext(asset, add_if_missing=True)

    assert dc.DatacubeExtension.has_extension(sample_item)
    assert dc.DatacubeExtension.get_schema_uri() in sample_item.stac_extensions


def test_apply(item: Item) -> None:
    dc.DatacubeExtension.add_to(item)
    dc.DatacubeExtension.ext(item).apply(
        dimensions={
            "time": dc.Dimension.from_dict(
                {"type": "temporal", "values": ["2016-05-03T13:21:30.040Z"]}
            ),
        },
        variables={
            "temp": dc.Variable.from_dict(
                {
                    "dimensions": [
                        "time",
                    ],
                }
            ),
        },
    )
    assert dc.DatacubeExtension.ext(item).dimensions
    assert dc.DatacubeExtension.ext(item).variables


def test_apply_without_required_fields_raises(item: Item) -> None:
    dc.DatacubeExtension.add_to(item)
    with pytest.raises(TypeError, match="missing 1 required positional argument"):
        dc.DatacubeExtension.ext(item).apply()  # type: ignore


@pytest.mark.vcr()
def test_validate(ext_item: Item) -> None:
    assert ext_item.validate()


def test_get_variables(ext_item: Item) -> None:
    prop = ext_item.properties[dc.VARIABLES_PROP]
    attr = dc.DatacubeExtension.ext(ext_item).variables

    assert attr is not None
    assert isinstance(attr["temp"], dc.Variable)
    assert attr["temp"].to_dict() == prop["temp"]


def test_get_dimensions(ext_item: Item) -> None:
    prop = ext_item.properties[dc.DIMENSIONS_PROP]
    attr = dc.DatacubeExtension.ext(ext_item).dimensions

    assert attr is not None
    assert isinstance(attr["spectral"], dc.Dimension)
    assert attr["spectral"].to_dict() == prop["spectral"]
    assert attr["x"].to_dict() == prop["x"]


@pytest.mark.vcr()
def test_set_variables(ext_item: Item) -> None:
    original = ext_item.properties[dc.VARIABLES_PROP]
    value = dc.Variable({"dimensions": ["foo"]})

    dc.DatacubeExtension.ext(ext_item).variables = {"temp": value}
    new = ext_item.properties[dc.VARIABLES_PROP]

    assert new != original
    assert new == {"temp": value.to_dict()}
    assert ext_item.validate()


@pytest.mark.vcr()
def test_set_dimensions(ext_item: Item) -> None:
    original = ext_item.properties[dc.DIMENSIONS_PROP]
    value = dc.Dimension(
        {"type": "bands", "values": ["red", "coastal", "green", "blue"]}
    )

    dc.DatacubeExtension.ext(ext_item).dimensions = {"spectral": value}
    new = ext_item.properties[dc.DIMENSIONS_PROP]

    assert new != original
    assert new == {"spectral": value.to_dict()}
    assert ext_item.validate()


@pytest.mark.parametrize("version", ["v1.0.0", "v2.0.0", "v2.1.0"])
def test_migrate(version: str, ext_item: Item) -> None:
    item_dict = ext_item.to_dict(include_self_link=False, transform_hrefs=False)
    item_dict["stac_extensions"] = [
        f"https://stac-extensions.github.io/datacube/{version}/schema.json"
    ]
    item = Item.from_dict(item_dict, migrate=True)
    assert (
        "https://stac-extensions.github.io/datacube/v2.2.0/schema.json"
        in item.stac_extensions
    )


def test_asset(item: Item) -> None:
    asset = Asset(href="data.tif")
    item.add_asset("data", asset)
    assert not asset.ext.has("cube")
    asset.ext.add("cube")
    assert asset.ext.has("cube")


def test_spatial_vector_object(item: Item) -> None:
    item.ext.add("cube")
    vector = dc.VectorSpatialDimension.from_dict(
        {
            "type": "geometries",
            "axes": ["x", "y"],
            "description": "A test vector spatial dimension",
            "bbox": [-180, -90, 180, 90],
            "values": ["POINT (-105.1019 40.1672)"],
            "geometry_types": ["Point"],
            "reference_system": 4326,
        }
    )
    item.ext.cube.dimensions = {"vector": vector}
    dimension = item.ext.cube.dimensions["vector"]
    assert isinstance(dimension, dc.VectorSpatialDimension)
    assert dimension.dim_type == "geometries"
    assert dimension.axes == ["x", "y"]
    assert dimension.description == "A test vector spatial dimension"
    assert dimension.bbox == [-180, -90, 180, 90]
    assert dimension.values == ["POINT (-105.1019 40.1672)"]
    assert dimension.geometry_types == ["Point"]
    assert dimension.reference_system == 4326



"""Tests for pystac.extensions.sar."""

from datetime import datetime
from random import choice
from string import ascii_letters

import pytest

import pystac
from pystac import ExtensionTypeError
from pystac.extensions import sar
from pystac.extensions.sar import (
    FrequencyBand,
    ObservationDirection,
    Polarization,
    SarExtension,
)
from pystac.summaries import RangeSummary
from tests.utils import TestCases


@pytest.fixture
def item() -> pystac.Item:
    asset_id = "my/items/2011"
    start = datetime(2020, 11, 7)
    item = pystac.Item(
        id=asset_id, geometry=None, bbox=None, datetime=start, properties={}
    )
    SarExtension.add_to(item)
    return item


@pytest.fixture
def sentinel_item() -> pystac.Item:
    return pystac.Item.from_file(TestCases.get_path("data-files/sar/sentinel-1.json"))


@pytest.fixture
def collection() -> pystac.Collection:
    return pystac.Collection.from_file(
        TestCases.get_path("data-files/collections/multi-extent.json")
    )


def test_stac_extensions(item: pystac.Item) -> None:
    assert SarExtension.has_extension(item)


@pytest.mark.vcr()
def test_required(item: pystac.Item) -> None:
    mode: str = "Nonsense mode"
    frequency_band: sar.FrequencyBand = sar.FrequencyBand.P
    polarizations: list[sar.Polarization] = [
        sar.Polarization.HV,
        sar.Polarization.VH,
    ]
    product_type: str = "Some product"

    SarExtension.ext(item).apply(mode, frequency_band, polarizations, product_type)
    assert mode == SarExtension.ext(item).instrument_mode
    assert sar.INSTRUMENT_MODE_PROP in item.properties

    assert frequency_band == SarExtension.ext(item).frequency_band
    assert sar.FREQUENCY_BAND_PROP in item.properties

    assert polarizations == SarExtension.ext(item).polarizations
    assert sar.POLARIZATIONS_PROP in item.properties

    assert product_type == SarExtension.ext(item).product_type
    assert sar.PRODUCT_TYPE_PROP in item.properties

    item.validate()


@pytest.mark.vcr()
def test_all(item: pystac.Item) -> None:
    mode: str = "WV"
    frequency_band: sar.FrequencyBand = sar.FrequencyBand.KA
    polarizations: list[sar.Polarization] = [
        sar.Polarization.VV,
        sar.Polarization.HH,
    ]
    product_type: str = "Some product"
    center_frequency: float = 1.2
    resolution_range: float = 3.1
    resolution_azimuth: float = 4.1
    pixel_spacing_range: float = 5.1
    pixel_spacing_azimuth: float = 6.1
    looks_range: int = 7
    looks_azimuth: int = 8
    looks_equivalent_number: float = 9.1
    observation_direction: sar.ObservationDirection = sar.ObservationDirection.LEFT

    SarExtension.ext(item).apply(
        mode,
        frequency_band,
        polarizations,
        product_type,
        center_frequency,
        resolution_range,
        resolution_azimuth,
        pixel_spacing_range,
        pixel_spacing_azimuth,
        looks_range,
        looks_azimuth,
        looks_equivalent_number,
        observation_direction,
    )

    assert center_frequency == SarExtension.ext(item).center_frequency
    assert sar.CENTER_FREQUENCY_PROP in item.properties

    assert resolution_range == SarExtension.ext(item).resolution_range
    assert sar.RESOLUTION_RANGE_PROP in item.properties

    assert resolution_azimuth == SarExtension.ext(item).resolution_azimuth
    assert sar.RESOLUTION_AZIMUTH_PROP in item.properties

    assert pixel_spacing_range == SarExtension.ext(item).pixel_spacing_range
    assert sar.PIXEL_SPACING_RANGE_PROP in item.properties

    assert pixel_spacing_azimuth == SarExtension.ext(item).pixel_spacing_azimuth
    assert sar.PIXEL_SPACING_AZIMUTH_PROP in item.properties

    assert looks_range == SarExtension.ext(item).looks_range
    assert sar.LOOKS_RANGE_PROP in item.properties

    assert looks_azimuth == SarExtension.ext(item).looks_azimuth
    assert sar.LOOKS_AZIMUTH_PROP in item.properties

    assert looks_equivalent_number == SarExtension.ext(item).looks_equivalent_number
    assert sar.LOOKS_EQUIVALENT_NUMBER_PROP in item.properties

    assert observation_direction == SarExtension.ext(item).observation_direction
    assert sar.OBSERVATION_DIRECTION_PROP in item.properties

    item.validate()


def test_polarization_must_be_list(item: pystac.Item) -> None:
    mode: str = "Nonsense mode"
    frequency_band: sar.FrequencyBand = sar.FrequencyBand.P
    # Skip type hint as we are passing in an incorrect polarization.
    polarizations = sar.Polarization.HV
    product_type: str = "Some product"
    with pytest.raises(pystac.STACError):
        SarExtension.ext(item).apply(
            mode,
            frequency_band,
            polarizations,  # type:ignore
            product_type,
        )


def test_should_return_none_when_observation_direction_is_not_set(
    item: pystac.Item,
) -> None:
    extension = SarExtension.ext(item)
    extension.apply(
        choice(ascii_letters),
        choice(list(sar.FrequencyBand)),
        [],
        choice(ascii_letters),
    )
    assert extension.observation_direction is None


def test_extension_not_implemented(sentinel_item: pystac.Item) -> None:
    # Should raise exception if Item does not include extension URI
    sentinel_item.stac_extensions.remove(SarExtension.get_schema_uri())

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = SarExtension.ext(sentinel_item)

    # Should raise exception if owning Item does not include extension URI
    asset = sentinel_item.assets["measurement"]

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = SarExtension.ext(asset)

    # Should succeed if Asset has no owner
    ownerless_asset = pystac.Asset.from_dict(asset.to_dict())
    _ = SarExtension.ext(ownerless_asset)


def test_item_ext_add_to(sentinel_item: pystac.Item) -> None:
    sentinel_item.stac_extensions.remove(SarExtension.get_schema_uri())
    assert SarExtension.get_schema_uri() not in sentinel_item.stac_extensions

    _ = SarExtension.ext(sentinel_item, add_if_missing=True)

    assert SarExtension.get_schema_uri() in sentinel_item.stac_extensions


def test_asset_ext_add_to(sentinel_item: pystac.Item) -> None:
    sentinel_item.stac_extensions.remove(SarExtension.get_schema_uri())
    assert SarExtension.get_schema_uri() not in sentinel_item.stac_extensions
    asset = sentinel_item.assets["measurement"]

    _ = SarExtension.ext(asset, add_if_missing=True)

    assert SarExtension.get_schema_uri() in sentinel_item.stac_extensions


def test_should_raise_exception_when_passing_invalid_extension_object() -> None:
    with pytest.raises(
        ExtensionTypeError,
        match=r"^SarExtension does not apply to type 'object'$",
    ):
        SarExtension.ext(object())  # type: ignore


def test_summaries_instrument_mode(collection: pystac.Collection) -> None:
    summaries_ext = SarExtension.summaries(collection, True)
    instrument_mode_list = ["WV"]

    summaries_ext.instrument_mode = instrument_mode_list

    assert summaries_ext.instrument_mode == instrument_mode_list
    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["sar:instrument_mode"] == instrument_mode_list


def test_summaries_frequency_band(collection: pystac.Collection) -> None:
    summaries_ext = SarExtension.summaries(collection, True)
    frequency_band_list = [FrequencyBand.P, FrequencyBand.L]

    summaries_ext.frequency_band = frequency_band_list

    assert summaries_ext.frequency_band == frequency_band_list
    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["sar:frequency_band"] == frequency_band_list


def test_summaries_polarizations(collection: pystac.Collection) -> None:
    summaries_ext = SarExtension.summaries(collection, True)
    polarizations_list = [Polarization.HH]

    summaries_ext.polarizations = polarizations_list

    assert summaries_ext.polarizations == polarizations_list

    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["sar:polarizations"] == polarizations_list


def test_summaries_product_type(collection: pystac.Collection) -> None:
    summaries_ext = SarExtension.summaries(collection, True)
    product_type_list = ["SSC"]

    summaries_ext.product_type = product_type_list

    assert summaries_ext.product_type == product_type_list

    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["sar:product_type"] == product_type_list


def test_summaries_center_frequency(collection: pystac.Collection) -> None:
    summaries_ext = SarExtension.summaries(collection, True)
    center_frequency_range = RangeSummary(4.405, 6.405)

    summaries_ext.center_frequency = center_frequency_range

    assert summaries_ext.center_frequency == center_frequency_range

    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["sar:center_frequency"] == center_frequency_range.to_dict()


def test_summaries_resolution_range(collection: pystac.Collection) -> None:
    summaries_ext = SarExtension.summaries(collection, True)
    resolution_range_range = RangeSummary(800.0, 1200.0)

    summaries_ext.resolution_range = resolution_range_range

    assert summaries_ext.resolution_range == resolution_range_range

    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["sar:resolution_range"] == resolution_range_range.to_dict()


def test_summaries_resolution_azimuth(collection: pystac.Collection) -> None:
    summaries_ext = SarExtension.summaries(collection, True)
    resolution_azimuth_range = RangeSummary(800.0, 1200.0)

    summaries_ext.resolution_azimuth = resolution_azimuth_range

    assert summaries_ext.resolution_azimuth == resolution_azimuth_range

    summaries_dict = collection.to_dict()["summaries"]

    assert (
        summaries_dict["sar:resolution_azimuth"] == resolution_azimuth_range.to_dict()
    )


def test_summaries_pixel_spacing_range(collection: pystac.Collection) -> None:
    summaries_ext = SarExtension.summaries(collection, True)
    pixel_spacing_range_range = RangeSummary(400.0, 600.0)

    summaries_ext.pixel_spacing_range = pixel_spacing_range_range

    assert summaries_ext.pixel_spacing_range == pixel_spacing_range_range

    summaries_dict = collection.to_dict()["summaries"]

    assert (
        summaries_dict["sar:pixel_spacing_range"] == pixel_spacing_range_range.to_dict()
    )


def test_summaries_pixel_spacing_azimuth(collection: pystac.Collection) -> None:
    summaries_ext = SarExtension.summaries(collection, True)
    pixel_spacing_azimuth_range = RangeSummary(400.0, 600.0)

    summaries_ext.pixel_spacing_azimuth = pixel_spacing_azimuth_range

    assert summaries_ext.pixel_spacing_azimuth == pixel_spacing_azimuth_range

    summaries_dict = collection.to_dict()["summaries"]

    assert (
        summaries_dict["sar:pixel_spacing_azimuth"]
        == pixel_spacing_azimuth_range.to_dict()
    )


def test_summaries_looks_range(collection: pystac.Collection) -> None:
    summaries_ext = SarExtension.summaries(collection, True)
    looks_range_range = RangeSummary(400, 600)

    summaries_ext.looks_range = looks_range_range

    assert summaries_ext.looks_range == looks_range_range

    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["sar:looks_range"] == looks_range_range.to_dict()


def test_summaries_looks_azimuth(collection: pystac.Collection) -> None:
    summaries_ext = SarExtension.summaries(collection, True)
    looks_azimuth_range = RangeSummary(400, 600)

    summaries_ext.looks_azimuth = looks_azimuth_range

    assert summaries_ext.looks_azimuth == looks_azimuth_range

    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["sar:looks_azimuth"] == looks_azimuth_range.to_dict()


def test_summaries_looks_equivalent_number(collection: pystac.Collection) -> None:
    summaries_ext = SarExtension.summaries(collection, True)
    looks_equivalent_number_range = RangeSummary(400.0, 600.0)

    summaries_ext.looks_equivalent_number = looks_equivalent_number_range

    assert summaries_ext.looks_equivalent_number == looks_equivalent_number_range

    summaries_dict = collection.to_dict()["summaries"]

    assert (
        summaries_dict["sar:looks_equivalent_number"]
        == looks_equivalent_number_range.to_dict()
    )


def test_summaries_observation_direction(collection: pystac.Collection) -> None:
    summaries_ext = SarExtension.summaries(collection, True)
    observation_direction_list = [ObservationDirection.LEFT]

    summaries_ext.observation_direction = observation_direction_list

    assert summaries_ext.observation_direction == observation_direction_list

    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["sar:observation_direction"] == observation_direction_list



import json
import os
import time
from datetime import datetime, timedelta, timezone

import pytest
from dateutil import tz

from pystac import utils
from pystac.utils import (
    JoinType,
    StringEnum,
    is_absolute_href,
    is_file_path,
    join_path_or_url,
    make_absolute_href,
    make_relative_href,
    now_in_utc,
    now_to_rfc3339_str,
    safe_urlparse,
    str_to_datetime,
)
from tests.utils import TestCases, path_includes_drive_letter


@pytest.mark.parametrize(
    "source_href, start_href, expected",
    (
        # relative href
        ("/a/b/c/d/catalog.json", "/a/b/c/catalog.json", "./d/catalog.json"),
        ("/a/b/catalog.json", "/a/b/c/catalog.json", "../catalog.json"),
        ("/a/catalog.json", "/a/b/c/catalog.json", "../../catalog.json"),
        ("/a/b/c/d/", "/a/b/c/catalog.json", "./d/"),
        ("/a/b/c/d/.dotfile", "/a/b/c/d/catalog.json", "./.dotfile"),
        (
            "file:///a/b/c/d/catalog.json",
            "file:///a/b/c/catalog.json",
            "./d/catalog.json",
        ),
        # relative href url
        (
            "http://stacspec.org/a/b/c/d/catalog.json",
            "http://stacspec.org/a/b/c/catalog.json",
            "./d/catalog.json",
        ),
        (
            "http://stacspec.org/a/b/catalog.json",
            "http://stacspec.org/a/b/c/catalog.json",
            "../catalog.json",
        ),
        (
            "http://stacspec.org/a/catalog.json",
            "http://stacspec.org/a/b/c/catalog.json",
            "../../catalog.json",
        ),
        (
            "http://stacspec.org/a/catalog.json",
            "http://cogeo.org/a/b/c/catalog.json",
            "http://stacspec.org/a/catalog.json",
        ),
        (
            "http://stacspec.org/a/catalog.json",
            "https://stacspec.org/a/b/c/catalog.json",
            "http://stacspec.org/a/catalog.json",
        ),
        (
            "http://stacspec.org/a/",
            "https://stacspec.org/a/b/c/catalog.json",
            "http://stacspec.org/a/",
        ),
        (
            "http://stacspec.org/a/b/.dotfile",
            "http://stacspec.org/a/b/catalog.json",
            "./.dotfile",
        ),
        # relative href under windows
        (
            "C:\\a\\b\\c\\d\\catalog.json",
            "C:\\a\\b\\c\\catalog.json",
            "./d/catalog.json",
        ),
        (
            "C:\\a\\b\\catalog.json",
            "C:\\a\\b\\c\\catalog.json",
            "../catalog.json",
        ),
        (
            "C:\\a\\catalog.json",
            "C:\\a\\b\\c\\catalog.json",
            "../../catalog.json",
        ),
        ("a\\b\\c\\catalog.json", "a\\b\\catalog.json", "./c/catalog.json"),
        ("a\\b\\catalog.json", "a\\b\\c\\catalog.json", "../catalog.json"),
    ),
)
def test_make_relative_href(source_href: str, start_href: str, expected: str) -> None:
    actual = make_relative_href(source_href, start_href)
    assert actual == expected


@pytest.mark.parametrize(
    "source_href, start_href, expected",
    (
        ("item.json", "/a/b/c/catalog.json", "/a/b/c/item.json"),
        ("./item.json", "/a/b/c/catalog.json", "/a/b/c/item.json"),
        ("./z/item.json", "/a/b/c/catalog.json", "/a/b/c/z/item.json"),
        ("../item.json", "/a/b/c/catalog.json", "/a/b/item.json"),
        (
            "item.json",
            "https://stacspec.org/a/b/c/catalog.json",
            "https://stacspec.org/a/b/c/item.json",
        ),
        (
            "./item.json",
            "https://stacspec.org/a/b/c/catalog.json",
            "https://stacspec.org/a/b/c/item.json",
        ),
        (
            "./z/item.json",
            "https://stacspec.org/a/b/c/catalog.json",
            "https://stacspec.org/a/b/c/z/item.json",
        ),
        (
            "../item.json",
            "https://stacspec.org/a/b/c/catalog.json",
            "https://stacspec.org/a/b/item.json",
        ),
        ("http://localhost:8000", None, "http://localhost:8000"),
        ("item.json", "file:///a/b/c/catalog.json", "file:///a/b/c/item.json"),
        (
            "./z/item.json",
            "file:///a/b/c/catalog.json",
            "file:///a/b/c/z/item.json",
        ),
        ("file:///a/b/c/item.json", None, "file:///a/b/c/item.json"),
        (
            "/vsigs/file.tif",
            "https://stacspec.org/a/b/c/catalog.json",
            "/vsigs/file.tif",
        ),
    ),
)
def test_make_absolute_href(source_href: str, start_href: str, expected: str) -> None:
    actual = make_absolute_href(source_href, start_href)
    if expected.startswith("file://"):
        _, actual = os.path.splitdrive(actual.replace("file://", ""))
        actual = f"file://{actual}"
    else:
        _, actual = os.path.splitdrive(actual)
    assert actual == expected


def test_make_absolute_href_on_vsitar() -> None:
    rel_path = "some/item.json"
    cat_path = "/vsitar//tmp/catalog.tar/catalog.json"
    expected = "/vsitar//tmp/catalog.tar/some/item.json"

    assert expected == make_absolute_href(rel_path, cat_path)


@pytest.mark.skipif(os.name != "nt", reason="Windows only test")
@pytest.mark.parametrize(
    "source_href, start_href, expected",
    (
        ("item.json", "C:\\a\\b\\c\\catalog.json", "C:/a/b/c/item.json"),
        (".\\item.json", "C:\\a\\b\\c\\catalog.json", "C:/a/b/c/item.json"),
        (
            ".\\z\\item.json",
            "Z:\\a\\b\\c\\catalog.json",
            "Z:/a/b/c/z/item.json",
        ),
        ("..\\item.json", "a:\\a\\b\\c\\catalog.json", "a:/a/b/item.json"),
        (
            "item.json",
            "HTTPS://stacspec.org/a/b/c/catalog.json",
            "https://stacspec.org/a/b/c/item.json",
        ),
        (
            "./item.json",
            "https://stacspec.org/a/b/c/catalog.json",
            "https://stacspec.org/a/b/c/item.json",
        ),
        (
            "./z/item.json",
            "https://stacspec.org/a/b/c/catalog.json",
            "https://stacspec.org/a/b/c/z/item.json",
        ),
        (
            "../item.json",
            "https://stacspec.org/a/b/c/catalog.json",
            "https://stacspec.org/a/b/item.json",
        ),
    ),
)
def test_make_absolute_href_windows(
    source_href: str, start_href: str, expected: str
) -> None:
    actual = make_absolute_href(source_href, start_href)
    assert actual == expected


def test_is_absolute_href() -> None:
    # Test cases of (href, expected, start_href)
    test_cases = [
        ("item.json", False, None),
        ("./item.json", False, None),
        ("../item.json", False, None),
        ("http://stacspec.org/item.json", True, None),
        ("http://stacspec.org/item.json", True, "http://stacspec.org/"),
        ("/vsigs/file.tiff", True, None),
        ("/vsigs/file.tiff", True, "http://stacspec.org/"),
    ]

    for href, expected, start_href in test_cases:
        actual = is_absolute_href(href, start_href)
        assert actual == expected


def test_is_absolute_href_os_aware() -> None:
    # Test cases of (href, expected)

    is_windows = os.name == "nt"
    incl_drive_letter = path_includes_drive_letter()
    test_cases = [
        ("/item.json", not incl_drive_letter, None),
        ("/item.json", False, "http://stacspec.org/"),
        ("/home/someuser/Downloads/item.json", not incl_drive_letter, None),
        ("file:///home/someuser/Downloads/item.json", not incl_drive_letter, None),
        ("d:/item.json", is_windows, None),
        ("c:/files/more_files/item.json", is_windows, None),
    ]

    for href, expected, start_href in test_cases:
        actual = is_absolute_href(href, start_href)
        assert actual == expected


@pytest.mark.skipif(os.name != "nt", reason="Windows only test")
def test_is_absolute_href_windows() -> None:
    # Test cases of (href, expected)

    test_cases = [
        ("item.json", False, None),
        (".\\item.json", False, None),
        ("..\\item.json", False, None),
        ("c:\\item.json", True, None),
        ("http://stacspec.org/item.json", True, None),
    ]

    for href, expected, start_href in test_cases:
        actual = is_absolute_href(href, start_href)
        assert actual == expected


def test_datetime_to_str() -> None:
    cases = (
        (
            "timezone naive, assume utc",
            datetime(2000, 1, 1),
            "2000-01-01T00:00:00Z",
        ),
        (
            "timezone aware, utc",
            datetime(2000, 1, 1, tzinfo=timezone.utc),
            "2000-01-01T00:00:00Z",
        ),
        (
            "timezone aware, utc -7",
            datetime(2000, 1, 1, tzinfo=timezone(timedelta(hours=-7))),
            "2000-01-01T00:00:00-07:00",
        ),
    )

    for title, dt, expected in cases:
        got = utils.datetime_to_str(dt)
        assert expected == got, f"Failure: {title}"


def test_datetime_to_str_with_microseconds_timespec() -> None:
    cases = (
        (
            "timezone naive, assume utc",
            datetime(2000, 1, 1, 0, 0, 0, 0),
            "2000-01-01T00:00:00.000000Z",
        ),
        (
            "timezone aware, utc",
            datetime(2000, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc),
            "2000-01-01T00:00:00.000000Z",
        ),
        (
            "timezone aware, utc -7",
            datetime(2000, 1, 1, 0, 0, 0, 0, tzinfo=timezone(timedelta(hours=-7))),
            "2000-01-01T00:00:00.000000-07:00",
        ),
    )

    for title, dt, expected in cases:
        got = utils.datetime_to_str(dt, timespec="microseconds")
        assert expected == got, f"Failure: {title}"


def test_str_to_datetime() -> None:
    def _set_tzinfo(tz_str: str | None) -> None:
        if tz_str is None:
            if "TZ" in os.environ:
                del os.environ["TZ"]
        else:
            os.environ["TZ"] = tz_str
        # time.tzset() only available for Unix/Linux
        if hasattr(time, "tzset"):
            time.tzset()

    utc_timestamp = "2015-06-27T10:25:31Z"

    prev_tz = os.environ.get("TZ")

    _set_tzinfo(None)
    utc_datetime = str_to_datetime(utc_timestamp)
    assert utc_datetime.tzinfo is tz.tzutc()
    assert utc_datetime.tzinfo is not tz.tzlocal()

    _set_tzinfo("UTC")
    utc_datetime = str_to_datetime(utc_timestamp)
    assert utc_datetime.tzinfo is tz.tzutc()
    assert utc_datetime.tzinfo is not tz.tzlocal()

    _set_tzinfo("US/Central")
    utc_datetime = str_to_datetime(utc_timestamp)
    assert utc_datetime.tzinfo is tz.tzutc()
    assert utc_datetime.tzinfo is not tz.tzlocal()

    if prev_tz is not None:
        _set_tzinfo(prev_tz)


def test_geojson_bbox() -> None:
    # Use sample Geojson from https://en.wikipedia.org/wiki/GeoJSON
    with open(
        TestCases.get_path("data-files/geojson/sample.geojson")
    ) as sample_geojson:
        all_features = json.load(sample_geojson)
        geom_dicts = [f["geometry"] for f in all_features["features"]]
        for geom in geom_dicts:
            got = utils.geometry_to_bbox(geom)
            assert got is not None


@pytest.mark.parametrize(
    "datetime",
    [
        "37-01-01T12:00:27.87Z",  # invalid year, must be 4 digits
        "21985-12-12T23:20:50.52Z",  # year must be 4 digits
        "1985-13-12T23:20:50.52Z",  # month > 12
        "1985-12-32T23:20:50.52Z",  # day > 31
        "1985-12-01T25:20:50.52Z",  # hour > 24
        "1985-12-01T00:60:50.52Z",  # minute > 59
        "1985-12-01T00:06:61.52Z",  # second > 60
        "1985-04-12T23:20:50.Z",  # fractional sec . but no frac secs
        "1985-04-12T23:20:50,Z",  # fractional sec , but no frac secs
        "1990-12-31T23:59:61Z",  # second > 60 w/o fractional seconds
    ],
)
def test_parse_invalid_rfc3339_str_to_datetime(datetime: str) -> None:
    with pytest.raises(ValueError):
        str_to_datetime(datetime)


@pytest.mark.parametrize(
    "datetime",
    [
        "1985-04-12",  # date only
        "1937-01-01T12:00:27.87+0100",  # invalid TZ format, no sep :
        "1985-12-12T23:20:50.52",  # no TZ
        "1985-04-12T23:20:50,52Z",  # comma as frac sec sep disallowed in RFC3339
    ],
)
def test_parse_invalid_rfc3339_but_valid_iso8601_str_to_datetime(datetime: str) -> None:
    assert str_to_datetime(datetime)


@pytest.mark.parametrize(
    "datetime",
    [
        "1985-04-12T23:20:50.52Z",
        "1996-12-19T16:39:57-00:00",
        "1996-12-19T16:39:57+00:00",
        "1996-12-19T16:39:57-08:00",
        "1996-12-19T16:39:57+08:00",
        "1937-01-01T12:00:27.87+01:00",
        "1985-04-12T23:20:50.52Z",
        "1937-01-01T12:00:27.8710+01:00",
        "1937-01-01T12:00:27.8+01:00",
        "1937-01-01T12:00:27.8Z",
        "2020-07-23T00:00:00.000+03:00",
        "2020-07-23T00:00:00+03:00",
        "1985-04-12t23:20:50.000z",
        "2020-07-23T00:00:00Z",
        "2020-07-23T00:00:00.0Z",
        "2020-07-23T00:00:00.01Z",
        "2020-07-23T00:00:00.012Z",
        "2020-07-23T00:00:00.0123Z",
        "2020-07-23T00:00:00.01234Z",
        "2020-07-23T00:00:00.012345Z",
        "2020-07-23T00:00:00.0123456Z",
        "2020-07-23T00:00:00.01234567Z",
        "2020-07-23T00:00:00.012345678Z",
    ],
)
def test_parse_valid_iso8601_str_to_datetime(datetime: str) -> None:
    assert str_to_datetime(datetime)


def test_now_functions() -> None:
    now1 = now_in_utc()
    time.sleep(1)
    now2 = now_in_utc()

    assert now1 < now2
    assert now1.tzinfo == timezone.utc

    assert str_to_datetime(now_to_rfc3339_str())


@pytest.mark.parametrize(
    "test_href,expected",
    [
        ("https://some/address/page.html", JoinType.URL),
        ("C:\\some\\windows\\path\\file.json", JoinType.PATH),
        ("some\\windows\\path\\file.json", JoinType.PATH),
        (".\\some\\windows\\path\\file.json", JoinType.PATH),
        ("C:/some/windows/path/file.json", JoinType.PATH),
        ("/some/posix/path/file.json", JoinType.PATH),
        ("./some/posix/path/file.json", JoinType.PATH),
        ("posix/path/file.json", JoinType.PATH),
    ],
)
def test_join_type(test_href: str, expected: JoinType) -> None:
    parsed = safe_urlparse(test_href)
    with pytest.warns(DeprecationWarning):
        assert JoinType.from_parsed_uri(parsed) == expected


def test_join_path_or_url() -> None:
    path_args = ["some", "path", "file.json"]
    with pytest.warns(DeprecationWarning):
        joined_path = join_path_or_url(JoinType.PATH, *path_args)
    if os.name != "nt":
        assert joined_path == "some/path/file.json"
    else:
        assert joined_path == "some\\path\\file.json"

    url_args = ["https://some", "page", "file.html"]
    with pytest.warns(DeprecationWarning):
        joined_url = join_path_or_url(JoinType.URL, *url_args)
    assert joined_url == "https://some/page/file.html"


@pytest.mark.parametrize(
    "href,expected",
    [
        ("path/to/file.txt", True),
        ("path/to/nofile", False),
        ("./path/to/file.txt", True),
        ("./path/to/nofile", False),
        ("./path/to/", False),
        ("/path/to/file.txt", True),
        ("/path/to/nofile", False),
        ("/path", False),
        ("/", False),
        ("D:/path/to/file.txt", True),
        ("D:/path/to/nofile", False),
        ("D:\\path\\to\\file.txt", True),
        ("D:\\path\\to\\file.txt", True),
        ("D:\\path\\to\\nofile", False),
        ("D:", False),
        ("D:\\", False),
        ("https://example.com/absolutepath/to/file.txt", True),
        ("https://example.com/absolutepath/to/nofile", False),
        ("https://example.com", False),
        ("https://example.com/", False),
    ],
)
def test_is_file_path(href: str, expected: bool) -> None:
    assert is_file_path(href) == expected


def test_stringenum_repr() -> None:
    class SomeEnum(StringEnum):
        THIS = "this"

    assert repr(SomeEnum.THIS) == "'this'"
